import os.path as op
from typing import List
import pandas as pd
import spacy
import transformers as trf
from datasets import Dataset, load_metric
from gramex.config import app_log
from gramex import cache
from sklearn.metrics import roc_auc_score


_CACHE = {}


def biluo2iob(tags: List[str]) -> List[str]:
    """Convert BILOU tags to IOB tags.

    spaCy insists on BILOU tags, but most transformers models use IOB tags.

    Parameters
    ----------
    tags : list
        List of BILOU tags

    Returns
    -------
    list
        List of IOB tags.

    Example
    -------
    >>> #      "Joe       R       Biden    is   President of  the  United   States   ."
    >>> tags = ['B-PER', 'I-PER', 'L-PER', 'O', 'U-PER', 'O', 'O', 'B-LOC', 'L-LOC', 'O']
    >>> biluo2iob(tags)
    ['B-PER', 'I-PER', 'I-PER', 'O', 'B-PER', 'O', 'O', 'B-LOC', 'I-LOC', 'O']
    """
    # Replace L
    tags = [t.replace("L-", "I-") for t in tags]
    # Replace U
    tags = [t.replace("U-", "B-") for t in tags]
    return tags


def offsets2iob(text: spacy.tokens.Doc, entities: List[dict]) -> List[str]:
    """Convert named entity offsets to a sequence of IOB tags.

    Parameters
    ----------
    text : spacy.tokens.Doc
        spaCy document of the original text
    entities : list
        Named entities present in the document as a list of dicts.
        Each dict represents one named entity and must contain three keys:
        1. "start": the start offset of the entity
        2. "end": the end offset of the entity
        3. "label": the label of the entity

    Returns
    -------
    list
        A list of IOB tags for the document.

    Example
    -------
    >>> import spacy
    >>> nlp = load('en')
    >>> doc = nlp('Narendra Modi is the PM of India.')
    >>> entities = [{'start': 0, 'end': 13, 'label': 'PER'},
    ...             {'start': 27, 'end': 32, 'label': 'LOC'}]
    >>> offsets2iob(doc, entities)
    ['B-PER', 'I-PER', 'O', 'O', 'O', 'O', 'B-LOC', 'O']
    """
    entities = [(ent["start"], ent["end"], ent["label"]) for ent in entities]
    tags = spacy.training.offsets_to_biluo_tags(text, entities)
    return biluo2iob(tags)


def tokenize_and_align_labels(examples, tokenizer):
    tokenized_inputs = tokenizer(examples["text"], truncation=True, is_split_into_words=True)

    labels = []
    for i, label in enumerate(examples["ner_tags"]):
        word_ids = tokenized_inputs.word_ids(batch_index=i)  # Map tokens to their respective word.
        previous_word_idx = None
        label_ids = []
        for word_idx in word_ids:  # Set the special tokens to -100.
            if word_idx is None:
                label_ids.append(-100)
            elif word_idx != previous_word_idx:  # Only label the first token of a given word.
                label_ids.append(label[word_idx])
            else:
                label_ids.append(-100)
            previous_word_idx = word_idx
        labels.append(label_ids)

    tokenized_inputs["labels"] = labels
    return tokenized_inputs


def load_pretrained(klass, path, default, **kwargs):
    if op.isdir(path):
        try:
            app_log.info(f"Attempting to load {klass.__name__} from {path}")
            model = cache.open(path, klass.from_pretrained, **kwargs)
        except Exception:
            app_log.info(f"Falling back to default {klass.__name__}: {default}.")
            model = cache.open(default, klass.from_pretrained, **kwargs)
    else:
        app_log.info(f"{path} not found on disk; loading default...")
        key = klass.__name__ + default
        if key in _CACHE:
            model = _CACHE[key]
        else:
            model = _CACHE[key] = klass.from_pretrained(default, **kwargs)
    return model


class BaseTransformer:
    def __init__(self, model=None, tokenizer=None, **kwargs):
        if model is None:
            model = self.DEFAULT_MODEL
        if tokenizer is None:
            tokenizer = self.DEFAULT_TOKENIZER
        self._model = model
        self._tokenizer = tokenizer
        self.model = load_pretrained(self.AUTO_CLASS, model, self.DEFAULT_MODEL)
        self.tokenizer = load_pretrained(trf.AutoTokenizer, tokenizer, self.DEFAULT_TOKENIZER)
        self.pipeline_kwargs = kwargs
        self.pipeline = trf.pipeline(
            self.task, model=self.model, tokenizer=self.tokenizer, **kwargs
        )

    def post_train(self, model_path):
        """Move the model to the CPU, save it with the tokenizer, recreate the pipeline."""
        self.model.to("cpu")
        self.model.save_pretrained(op.join(model_path, "model"))
        self.tokenizer.save_pretrained(op.join(model_path, "tokenizer"))
        self.pipeline = trf.pipeline(
            self.task,
            model=self.model,
            tokenizer=self.tokenizer,
            **self.pipeline_kwargs,
        )


class SentimentAnalysis(BaseTransformer):
    task = "sentiment-analysis"
    DEFAULT_MODEL = DEFAULT_TOKENIZER = "distilbert-base-uncased-finetuned-sst-2-english"
    AUTO_CLASS = trf.AutoModelForSequenceClassification

    def fit(self, text, labels, model_path, **kwargs):
        if pd.api.types.is_object_dtype(labels):
            labels = labels.map(self.model.config.label2id.get)
        ds = Dataset.from_dict({"text": text, "label": labels})
        tokenized = ds.map(
            lambda x: self.tokenizer(x["text"], padding="max_length", truncation=True),
            batched=True,
        )
        train_args = trf.TrainingArguments(save_strategy="no", output_dir=model_path)
        trainer = trf.Trainer(model=self.model, train_dataset=tokenized, args=train_args)
        trainer.train()
        self.post_train(model_path)

    def predict(self, text, **kwargs):
        text = text.tolist()
        predictions = self.pipeline(text)
        return [k["label"] for k in predictions]

    def score(self, X, y_true, **kwargs):  # noqa N803
        y_true = [self.model.config.label2id[x] for x in y_true]
        y_pred = self.predict(X.squeeze("columns"))
        y_pred = [self.model.config.label2id[x] for x in y_pred]
        return roc_auc_score(y_true, y_pred)


class NER(BaseTransformer):
    task = "ner"
    DEFAULT_TOKENIZER = DEFAULT_MODEL = "dbmdz/bert-large-cased-finetuned-conll03-english"
    AUTO_CLASS = trf.AutoModelForTokenClassification

    def __init__(self, model=None, tokenizer=None, **kwargs):
        self.nlp = spacy.blank("en")
        super(NER, self).__init__(
            model=model, tokenizer=tokenizer, aggregation_strategy="first", **kwargs
        )

    @property
    def labels(self):
        return {k.split("-")[-1] for k in self.model.config.label2id}

    def predict(self, text, **kwargs):
        text = text.tolist()
        return self.pipeline(text)

    def score(self, X, y_true, **kwargs):
        try:
            metric = load_metric("seqeval")
        except ImportError:
            app_log.error("Could not load the seqeval metric. Scoring not supported.")
            return 0
        # Get references and predictions
        X = X.squeeze("columns")
        predictions = self.predict(X)
        for pred in predictions:
            for ent in pred:
                ent.update({"label": ent.pop("entity_group")})
        preds = []
        refs = []
        for doc, pred, ref in zip(self.nlp.pipe(X.tolist()), predictions, y_true):
            preds.append(offsets2iob(doc, pred))
            refs.append(offsets2iob(doc, ref))
        metrics = metric.compute(references=refs, predictions=preds)
        return pd.DataFrame({k: v for k, v in metrics.items() if k in self.labels}).reset_index()

    def fit(self, text, labels, model_path, **kwargs):
        texts = []
        ner_tags = []
        for doc, ents in zip(self.nlp.pipe(text.tolist()), labels):
            texts.append([t.text for t in doc])
            ner_tags.append(offsets2iob(doc, ents))

        label2id = self.model.config.label2id
        ner_tags = [[label2id.get(k, 0) for k in tags] for tags in ner_tags]

        dataset = Dataset.from_dict({"text": texts, "ner_tags": ner_tags})
        tokenized = dataset.map(
            lambda x: tokenize_and_align_labels(x, self.tokenizer), batched=True
        )
        collator = trf.DataCollatorForTokenClassification(tokenizer=self.tokenizer)
        args = trf.TrainingArguments(
            save_strategy="no", output_dir=model_path, evaluation_strategy="epoch"
        )
        trainer = trf.Trainer(
            model=self.model,
            args=args,
            train_dataset=tokenized,
            eval_dataset=tokenized,
            tokenizer=self.tokenizer,
            data_collator=collator,
        )
        trainer.train()
        self.post_train(model_path)



import gramex.cache
from . import TestGramex
from nose.tools import eq_, ok_


class TestOpenAPIHandler(TestGramex):
    expected = gramex.cache.open('openapiresponse.yaml', rel=True)

    def has_param(self, params, **kwargs):
        items = kwargs.items()
        return any(items <= param.items() for param in params)

    def test_openapi(self):
        # OpenAPI spec is a JSON response
        spec = self.check('/openapi/spec').json()
        # OpenAPI spec version matches
        eq_(spec['openapi'], '3.0.2')
        # spec.info comes from gramex.yaml OpenAPIHandler kwargs
        # .info is from gramex.yaml > openapi/spec
        eq_(
            spec['info'],
            {
                'title': 'OpenAPI-title',
                'version': 'OpenAPI-version',
                'description': 'OpenAPI-description',
            },
        )
        # spec.servers comes from gramex.yaml OpenAPIHandler kwargs
        eq_(spec['servers'], [{'url': '..', 'description': 'Server-description'}])

        self.check_functionhandler(spec)
        self.check_formhandler(spec)

    def check_functionhandler(self, spec):
        # /openapi/func path exists
        ok_('/openapi/func' in spec['paths'])
        path = spec['paths']['/openapi/func']
        for request in ('post', 'put'):
            ok_(request in path)
            conf = path[request]
            # Summary is based on function name
            eq_(conf['summary'], 'Openapi Func: FunctionHandler')
            # Description is as per utils.test_function
            eq_(conf['description'], '\nThis is a **Markdown** docstring.\n')
            # Argument types, defaults, required as per utils.test_function
            params = {param['name']: param for param in conf['parameters']}
            self.assertDictContainsSubset(
                {
                    'in': 'query',
                    'description': '',
                    'required': True,
                    'schema': {'type': 'array', 'items': {'type': 'integer'}},
                },
                params['li1'],
            )
            self.assertDictContainsSubset(
                {
                    'in': 'query',
                    'description': '',
                    'required': True,
                    'schema': {'type': 'array', 'items': {'type': 'number'}},
                },
                params['lf1'],
            )
            self.assertDictContainsSubset(
                {
                    'in': 'query',
                    'description': 'List of ints',
                    'required': True,
                    'schema': {'type': 'array', 'items': {'type': 'integer'}},
                },
                params['li2'],
            )
            self.assertDictContainsSubset(
                {
                    'in': 'query',
                    'description': 'List of floats',
                    'required': True,
                    'schema': {'type': 'array', 'items': {'type': 'number'}},
                },
                params['lf2'],
            )
            self.assertDictContainsSubset(
                {
                    'in': 'query',
                    'description': '',
                    'schema': {'type': 'array', 'items': {'type': 'integer'}, 'default': [0]},
                },
                params['li3'],
            )
            self.assertDictContainsSubset(
                {
                    'in': 'query',
                    'description': '',
                    'schema': {'type': 'array', 'items': {'type': 'number'}, 'default': [0.0]},
                },
                params['lf3'],
            )
            self.assertDictContainsSubset(
                {
                    'in': 'query',
                    'description': '',
                    'schema': {'type': 'string', 'default': []},
                },
                params['l1'],
            )
            self.assertDictContainsSubset(
                {
                    'in': 'query',
                    'description': 'First value',
                    'schema': {'type': 'integer', 'default': 0},
                },
                params['i1'],
            )
            self.assertDictContainsSubset(
                {
                    'in': 'query',
                    'description': 'Second value',
                    'schema': {'type': 'integer', 'default': 0},
                },
                params['i2'],
            )
            self.assertDictContainsSubset(
                {
                    'in': 'query',
                    'description': '',
                    'schema': {'type': 'string', 'default': 'Total'},
                },
                params['s1'],
            )
            self.assertDictContainsSubset(
                {
                    'in': 'query',
                    'description': '',
                    'schema': {'type': 'integer', 'default': 0},
                },
                params['n1'],
            )
            self.assertDictContainsSubset(
                {
                    'in': 'query',
                    'description': '',
                    'schema': {'type': 'string', 'default': 0},
                },
                params['n2'],
            )
            self.assertDictContainsSubset(
                {
                    'in': 'header',
                    'description': '',
                    'schema': {'type': 'string', 'default': ''},
                },
                params['h'],
            )
            self.assertDictContainsSubset(
                {
                    'in': 'query',
                    'description': '',
                    'schema': {'type': 'integer', 'default': 200},
                },
                params['code'],
            )

            resp = conf['responses']
            # 400 response has description from gramex.yaml
            self.assertDictContainsSubset(
                {
                    'description': 'Bad request',
                    'content': {'text/html': {'example': 'Bad request'}},
                },
                resp['400'],
            )
            # Rest should have default error responses
            self.assertDictContainsSubset(
                {'description': 'Successful Response', 'content': {'application/json': {}}},
                resp['200'],
            )
            self.assertDictContainsSubset(
                {
                    'description': 'Not authorized',
                    'content': {'text/html': {'example': 'Not authorized'}},
                },
                resp['401'],
            )
            # 429 is added only to PUT not GET
            if request == 'put':
                self.assertDictContainsSubset({'description': 'Rate limited'}, resp['429'])
            else:
                self.assertNotIn('429', resp)

    def check_formhandler(self, spec):
        ok_('/openapi/form' in spec['paths'])
        eq_(spec['paths']['/openapi/form'], self.expected['/openapi/form'])
        eq_(spec['paths']['/openapi/form'], self.expected['/openapi/form'])



import ast
import datetime
import importlib
import json
import os
import time
import tornado.gen
import yaml
from functools import wraps
from types import GeneratorType
from orderedattrdict import AttrDict
from gramex.config import app_log, locate, variables, CustomJSONEncoder, merge
from typing import Union, Any, List


def identity(x):
    return x


def _arg_repr(arg):
    '''
    Arguments starting with `=` are converted into the variable. Otherwise,
    values are treated as strings. For example, `=x` is the variable `x` but
    `x` is the string `"x"`. `==x` is the string `"=x"`.
    '''
    if isinstance(arg, str):
        if arg.startswith('=='):
            return repr(arg[1:])  # "==x" becomes '"=x"'
        elif arg.startswith('='):
            return arg[1:]  # "=x" becomes 'x'
    return repr(arg)  # "x" becomes '"x"', 1 becomes '1', etc


def _full_name(tree):
    '''Decompile ast tree for "x", "module.x", "package.module.x", etc'''
    if isinstance(tree, ast.Name):
        return tree.id
    elif isinstance(tree, ast.Attribute):
        parent = _full_name(tree.value)
        return parent + '.' + tree.attr if parent is not None else parent
    return None


def module_names(node, vars):
    '''
    Collects a list of modules mentioned in an AST tree. Ignores things in vars

    visitor = ModuleNameVisitor()
    visitor.visit(ast.parse(expression))
    visitor.modules
    '''
    context = []
    modules = set()

    def visit(node):
        if not hasattr(node, '_fields'):
            return
        for child in ast.iter_child_nodes(node):
            if isinstance(child, ast.Name) and len(context) and context[-1]:
                module = [child.id]
                for p in context[::-1]:
                    if p is not None:
                        module.append(p)
                    else:
                        break
                if len(module) and module[0] not in vars:
                    module.pop()
                    while len(module):
                        module_name = '.'.join(module)
                        try:
                            importlib.import_module(module_name)
                            modules.add(module_name)
                            break
                        except ImportError:
                            module.pop()
                        # Anything other than an ImportError means we've identified the module.
                        # E.g. A SyntaxError means the file is right, it just has an error.
                        # Add these modules as well.
                        else:
                            modules.add(module_name)
                            break
            context.append(child.attr if isinstance(child, ast.Attribute) else None)
            visit(child)
            context.pop()

    visit(node)
    return modules


def build_transform(
    conf: dict,
    vars: dict = None,
    kwargs: Union[bool, str] = False,
    filename: str = 'transform',
    cache: bool = False,
    iter: bool = True,
):
    '''Converts an expression into a callable function.

    Examples:
        >>> fn = build_transform({'function': 'json.dumps({"x": 1})'}, iter=False)
        >>> fn()
        ... '{"x": 1}'

    This compiles the expression into a callable function as follows:

    ```python
    def transform(_val):
        import json
        result = json.dumps({"x": 1})
        return result
    ```

    Parameters:
        conf: expression to compile
        vars: variables passed to the compiled function
        kwargs: `True` to accept **kwargs. Any string to define the kwargs variable name
        filename: filename to print in case of errors
        cache: `False` re-imports modules if changed. `True` caches module
        iter: `True` always returns an iterable. `False` returns a single value

    `conf` is a dict with a `function` key with Python expression. This expression is compiled
    and returned. If the expression uses modules (e.g. `json.dumps`), they are auto-imported.

    It optionally accepts an `args` list and a `kwargs` list if `function` is a function name.

    Examples:
        >>> {'function': '1'}               # => 1
        >>> {'function': 'x + y'}           # => x + y
        >>> {'function': 'json.dumps(x)'}   # => json.dumps(s)
        >>> {'function': 'json.dumps', 'args': ['x'], 'kwargs': {'indent': 2}}
        >>> {'function': 'json.dumps("x", indent=2)`    # same as above

    `args` values and `kwargs` keys are treated as strings, not variables. But values starting with
    `=` (e.g. `=handler`) are treated as variables. (Use `==x` to represent the string `=x`.)

    Examples:
        >>> {'function': 'str', 'args': ['handler']}      # => str('handler')
        >>> {'function': 'str', 'args': ['=handler']}     # => str(handler)
        >>> {'function': 'str', 'args': ['==handler']}    # => str('=handler')

    `vars` defines the compiled function's signature. `vars={'x': 1, 'y': 2}` creates a
    `def transform(x=1, y=2):`.

    `vars` defaults to `{'_val': None}`.

    Examples:
        >>> add = build_transform({'function': 'x + y'}, vars={'x': 1, 'y': 2}, iter=False)
        >>> add()
        ... 3
        >>> add(x=3)
        ... 5
        >>> add(x=3, y=4)
        ... 7
        >>> incr = build_transform({'function': '_val + 1'}, iter=False)
        >>> incr(3)
        ... 4

    `kwargs=True` allows the compiled function to accept any keyword arguments as `**kwargs`.
    Specify `kwargs='kw'` to use `kw` (or any string) as the keyword arguments variable instead.

    Examples:
        >>> params = build_transform({'function': 'kwargs'}, vars={}, kwargs=True, iter=False)
        >>> params(x=1)
        ... {'x': 1}
        >>> params(x=1, y=2)
        ... {'x': 1, 'y': 2}
        >>> params = build_transform({'function': 'kw'}, vars={}, kwargs='kw', iter=False)
        >>> params(x=1, y=2)
        ... {'x': 1, 'y': 2}

    `filename` defines the filename printed in error messages.

    Examples:
        >>> error = build_transform({'function': '1/0'}, filename='zero-error', iter=False)
        ... Traceback (most recent call last):
        ...   File "<stdin>", line 1, in <module>
        ...   File "zero-error", line 2, in transform
        ... ZeroDivisionError: division by zero

    `cache=False` re-imports the modules if changed. This is fairly efficient, and is the default.
    Use `cache=True` to cache modules until Python is restarted.

    `iter=True` always returns an iterable. If the `function` is a generator (i.e. has a `yield`),
    it is returned as-is. Else it is returned as an array, i.e. `[result]`.

    Examples:
        `build_transform()` returns results wrapped as an array.

        >>> val = build_transform({'function': '4'})
        >>> val()
        ... [4]
        >>> val = build_transform({'function': '[4, 5]'})
        >>> val()
        ... [[4, 5]]

        If the result is a generator, it is returned as-is.

        >>> def gen():
        ...     for x in range(5):
        ...         yield x
        >>> val = build_transform({'function': 'fn()'}, vars={'fn': None})
        >>> val(gen)
        ... <generator object gen at 0x....>
        >>> list(val(gen))
        ... [0, 1, 2, 3, 4]

        If `iter=False`, it returns the results as-is.

        >>> val = build_transform({'function': '4'}, iter=False)
        >>> val()
        ... 4
        >>> val = build_transform({'function': '[4, 5]'}, iter=False)
        >>> val()
        ... [4, 5]
    '''
    # Ensure that the transform is a dict with "function:" in it. (This is a common mistake)
    if not isinstance(conf, dict) or 'function' not in conf:
        raise ValueError(f'{filename}: needs "function:". Got {conf!r}')

    conf = {key: val for key, val in conf.items() if key in {'function', 'args', 'kwargs'}}

    # The returned function takes a single argument by default
    if vars is None:
        vars = {'_val': None}
    # Treat kwargs=True as kwargs=kwargs. It adds **kwargs to the function call
    if kwargs is True:
        kwargs = 'kwargs'

    # If the function is a list, treat it as a pipeline
    if isinstance(conf['function'], (list, tuple)):
        return build_pipeline(conf['function'], vars, kwargs, filename, cache, iter)

    # Get the name of the function in case it's specified as a function call
    # expr is the full function / expression, e.g. str("abc")
    # tree is the ast result
    expr = str(conf['function'])
    tree = ast.parse(expr)
    if len(tree.body) != 1 or not isinstance(tree.body[0], ast.Expr):
        raise ValueError(f'{filename}: function: must be Python function or expr, not {expr}')

    # Check whether to use the expression as is, or construct the expression
    # If expr is like "x" or "module.x", construct it if it's callable
    # Else, use the expression as-is
    function_name = _full_name(tree.body[0].value)
    module_name = function_name.split('.')[0] if isinstance(function_name, str) else None
    function, doc = None, expr
    # If the module or function is one of the vars themselves, return it as-is
    # _val.type will be used as-is, then, rather than looking for an "_val" module
    if module_name in vars or (isinstance(kwargs, str) and module_name == kwargs):
        expr = function_name
    # If it's a function call, construct the function signature
    elif function_name is not None:
        function = locate(function_name, modules=['gramex.transforms'])
        doc = function.__doc__
        if function is None:
            app_log.error(f'function:{filename}: Cannot load function {function_name}')
        # This section converts the function into an expression.
        # We do this only if the original expression was a *callable* function.
        # But if we can't load the original function (e.g. SyntaxError),
        # treat that as a function as well, allowing users to correct it later.
        if callable(function) or function is None:
            if 'args' in conf:
                # If args is not a list, convert to a list with that value
                args = conf['args'] if isinstance(conf['args'], list) else [conf['args']]
            else:
                # If args is not specified, use vars' keys as args
                args = [f'={var}' for var in vars.keys()]
            # Add the function, arguments, and kwargs
            expr = function_name + '('
            for arg in args:
                expr += f'{_arg_repr(arg)}, '
            for key, val in conf.get('kwargs', {}).items():
                expr += f'{key}={_arg_repr(val)}, '
            expr += ')'
    # If expr starts with a function call (e.g. module.func(...)), use it's docs
    elif isinstance(tree.body[0].value, ast.Call):
        from astor import to_source

        doc = locate(to_source(tree.body[0].value.func).strip()).__doc__

    # Create the code
    modules = module_names(tree, vars)
    modulestr = ', '.join(sorted(modules))
    signature = ', '.join('{:s}={!r:}'.format(k, v) for k, v in vars.items())
    if kwargs:
        signature += (', ' if signature else '') + f'**{kwargs}'
    body = [
        f'def transform({signature}):\n',
        f'\timport {modulestr}\n' if modulestr else '',
        f'\treload_module({modulestr})\n' if modulestr and not cache else '',
        f'\tresult = {expr}\n',
        # If the result is a generator object, return it. Else, create a list and
        # return that. This ensures that the returned value is always an iterable
        (
            '\treturn result if isinstance(result, GeneratorType) else [result,]'
            if iter
            else '\treturn result'
        ),
    ]

    # Compile the function with context variables
    import gramex.transforms
    from gramex.cache import reload_module

    context = dict(
        reload_module=reload_module,
        GeneratorType=GeneratorType,
        Return=tornado.gen.Return,
        AttrDict=AttrDict,
        **{key: getattr(gramex.transforms, key) for key in gramex.transforms.__all__},
    )
    code = compile(''.join(body), filename=filename, mode='exec')
    # B102:exec_used is safe since the code is written by app developer
    exec(code, context)  # noqa S102

    # Return the transformed function
    result = context['transform']
    result.__name__ = str(function_name or filename)
    result.__doc__ = str(doc)
    result.__func__ = function

    return result


def build_pipeline(
    conf: dict,
    vars: dict = None,
    kwargs: Union[bool, str] = False,
    filename: str = 'pipeline',
    cache: bool = False,
    iter: bool = True,
):
    '''Converts an expression list into a callable function (called a pipeline).

    Examples:
        >>> fn = build_pipeline([
        ...     {'name': 'x', 'function': '1 + 2'},
        ...     {'name': 'y', 'function': '3 + 4'},
        ...     {'function': 'x + y'},
        ... ], iter=False)
        >>> fn()
        ... 10

    This compiles the expression list into a callable function roughty as follows:

    ```python
    def pipeline(_val):
        x = 1 + 2
        y = 3 + 4
        return x + y
    ```

    Parameters:
        conf: expression list to compile
        vars: variables passed to the compiled function
        kwargs: `True` to accept **kwargs. Any string to define the kwargs variable name
        filename: filename to print in case of errors
        cache: `False` re-imports modules if changed. `True` caches module
        iter: `True` always returns an iterable. `False` returns a single value

    `conf` is a **list** of the same `conf` that
    [build_transform][gramex.transforms.build_transform] accepts.

    Other parameters are the same as [build_transform][gramex.transforms.build_transform].
    '''
    if not isinstance(conf, (list, tuple)):
        raise ValueError(f'pipeline:{filename}: must be a list, not {type(conf)}')
    if len(conf) == 0:
        raise ValueError(f'pipeline:{filename}: cannot be an empty list')
    if vars is None:
        vars = {}
    if not isinstance(vars, dict):
        raise ValueError(f'pipeline:{filename}: vars must be a dict, not {type(vars)}')
    # current_scope has the variables available in each stage.
    # Whenever a stage defines a `name:`, add it to the current_scope.
    current_scope = dict(vars)
    n, compiled_stages = len(conf), []
    for index, spec in enumerate(conf, start=1):
        if not isinstance(spec, dict):
            spec = {'function': str(spec)}
        # Store the original configuration for reporting error messages
        stage = {'spec': spec, 'index': index}
        if 'function' not in spec:
            raise ValueError(f'pipeline:{filename}: {index}/{n}: missing "function"')
        # Compile the function, allowing use of all variables in current_scope
        stage['function'] = build_transform(
            {'function': spec['function']},
            vars=current_scope,
            kwargs=kwargs,
            filename=f'pipeline:{filename} {index}/{n}',
            cache=cache,
            iter=False,
        )
        # If the stage defines a name, add it as a variable for current_scope
        if 'name' in spec:
            current_scope[spec['name']] = None
        compiled_stages.append(stage)

    def run_pipeline(**kwargs):
        '''
        This returned function actually runs the pipeline.
        It loops through each pipeline step, runs the function, and returns the last value.

        Any `kwargs` passed are used as globals. (They default to the `vars` in build_pipeline.)

        If a step specifies a `name`, the result is stored in `kwargs[name]`,
        making it available as a global to the next step.

        Logs the time taken for each step (and errors, if any) in storelocations.pipeline.
        '''
        start = datetime.datetime.utcnow().isoformat()
        app_log.debug(f'pipeline:{filename} running')
        error, stage = '', {'spec': {}, 'index': None}
        try:
            # Use kwargs as globals for the steps. Initialize with vars
            merge(kwargs, vars, 'setdefault')
            result = None
            for stage in compiled_stages:
                result = stage['function'](**kwargs)
                # Store the returned value in the kwargs as globals
                if 'name' in stage['spec']:
                    kwargs[stage['spec']['name']] = result
            # If the pipeline SHOULD return an iterable, ensure last result is iterable
            if iter:
                return result if isinstance(result, GeneratorType) else [result]
            else:
                return result
        except Exception:
            # On any exception, capture the error and traceback to log it
            import sys
            import traceback

            error = f'pipeline:{filename} {stage["index"]}/{n} failed: {stage["spec"]}'
            error += '\n' + ''.join(traceback.format_exception(*sys.exc_info()))
            # but raise the original Exception
            raise
        finally:
            # Log pipeline execution (and error, if any)
            from gramex.services import info

            if 'pipeline' in info.storelocations:
                from gramex.data import insert

                end = datetime.datetime.utcnow().isoformat()
                try:
                    insert(
                        **info.storelocations.pipeline,
                        id=['name', 'start'],
                        args={
                            'name': [filename],
                            'start': [start],
                            'end': [end],
                            'error': [error],
                        },
                    )
                except Exception:
                    app_log.exception(f'pipeline:{filename} exception logging failed')

    return run_pipeline


def condition(*args):
    '''
    !!! Deprecated
        Use the `if` construct in config keys instead.

    Variables can also be computed based on conditions:

        variables:
            OS:
                default: 'No OS variable defined'
            PORT:
                function: condition
                args:
                    - $OS.startswith('Windows')
                    - 9991
                    - $OS.startswith('Linux')
                    - 9992
                    - 8883
    '''
    import warnings

    warnings.warn(
        'condition() deprecated. https://gramener.com/gramex/guide/config/#conditions',
        DeprecationWarning,
    )
    from string import Template

    var_defaults = {}
    for var in variables:
        var_defaults[var] = f"variables.get('{var}', '')"
    # could use iter, range(.., 2)
    if len(args) == 1 and isinstance(args[0], dict):
        pairs = args[0].items()
    else:
        pairs = zip(args[0::2], args[1::2])
    for cond, val in pairs:
        if isinstance(cond, str):
            # B307:eval is safe here since `cond` is written by app developer
            if eval(Template(cond).substitute(var_defaults)):  # noqa S307
                return val
        elif cond:
            return val

    # If none of the conditions matched, we'll be here.
    # If there are an odd number of arguments and there's at least one condition,
    # treat the last as a default.
    if len(args) % 2 == 1 and len(args) > 2:
        return args[-1]


def flattener(fields: dict, default: Any = None, filename: str = 'flatten'):
    '''Generates a function that flattens deep dictionaries.

    Examples:
        >>> flat = flattener({
        ...     'id': 'id',
        ...     'name': 'user.screen_name'
        ... })
        >>> flat({'id': 1, 'user': {'screen_name': 'name'}})
        ... {'id': 1, 'name': 'name'}

    Parameters:
        fields: a mapping of keys to object paths
        default: the value to use if the object path is not found
        filename: filename to print in case of errors

    Object paths are dot-seperated, and constructed as follows:

    ```text
    '1'   => obj[1]
    'x'   => obj['x']
    'x.y' => obj['x']['y']
    '1.x' => obj[1]['x']
    ```
    '''
    body = [
        f'def {filename}(obj):\n',
        '\tr = AttrDict()\n',
    ]

    def assign(field, target, catch_errors=False):
        if catch_errors:
            body.append(f'\ttry: r[{field!r}] = {target}\n')
            body.append(f'\texcept (KeyError, TypeError, IndexError): r[{field!r}] = default\n')
        else:
            body.append(f'\tr[{field!r}] = {target}\n')

    for field, source in fields.items():
        if not isinstance(field, str):
            app_log.error(f'flattener:{filename}: key {field} is not a str')
            continue
        if isinstance(source, str):
            target = 'obj'
            if source:
                for item in source.split('.'):
                    target += ('[%s]' if item.isdigit() else '[%r]') % item
            assign(field, target, catch_errors=True)
        elif source is True:
            assign(field, 'obj')
        elif isinstance(source, int) and not isinstance(source, bool):
            assign(field, f'obj[{source}]', catch_errors=True)
        else:
            app_log.error(f'flattener:{filename}: value {source} is not a str/int')
            continue
    body.append('\treturn r')
    code = compile(''.join(body), filename=f'flattener:{filename}', mode='exec')
    context = {'AttrDict': AttrDict, 'default': default}
    # B307:eval is safe here since the code is constructed entirely in this function
    eval(code, context)  # noqa S307
    return context[filename]


_once_info = {}


def once(*args, **kwargs):
    '''Returns `False` if once() has been called before with these arguments. Else `True`.

    Data is stored in a persistent SQLite dict.
    '''
    if 'db' not in _once_info:
        import os
        from sqlitedict import SqliteDict

        dbpath = os.path.join(variables['GRAMEXDATA'], 'once.db')
        _once_info['db'] = SqliteDict(dbpath, tablename='once', autocommit=True)
    db = _once_info['db']
    key = json.dumps(args, separators=(',', ':'), cls=CustomJSONEncoder)
    if kwargs.get('_clear', False):
        if key in db:
            del db[key]
        return None
    if key in db:
        return False
    db[key] = True
    return True


# int(x), float(x), str(x) do a good job of converting strings to respective types.
# But not all types work smoothly. Handle them here.
_convert_map = {
    # bool("true") fails Use yaml.safe_load in such cases
    bool: lambda x: yaml.safe_load(x) if isinstance(x, (str, bytes)) else x,
    # NoneType("None") doesn't work either. Just return None
    type(None): lambda x: None,
    # TODO: Convert dates but without importing pandas on startup
    # datetime.datetime: lambda x: pd.to_datetime(x).to_pydatetime,
}


def typelist(hint):
    typ, is_list = hint, False
    # If typ is an Annotation, use the native type. Else use the type.
    while hasattr(typ, '__args__'):
        is_list = getattr(typ, '_name', None) in ('List', 'Tuple')
        typ = typ.__args__[0]
    return typ, is_list


def convert(hint, param, *args):
    from pandas.core.common import flatten

    args = list(flatten(args))
    typ, is_list = typelist(hint)
    # Convert args to the native type
    method = _convert_map.get(typ, typ)
    # If default is list-like or hint is List-like, return list. Else return last value
    if is_list or isinstance(param.default, (list, tuple)):
        return [method(arg) for arg in args]
    else:
        return method(args[-1] if args else param.default)


class Header(str):
    pass


def handler(func):
    """Wrap a function to make it compatible with a FunctionHandler.

    Use this decorator to expose a function as a REST API, with path or URL parameters mapped to
    function arguments with type conversion.

    Suppose you have the following function in `greet.py`:

    ```python
    @handler
    def birthday(name: str, age: int):
        return f'{name} turns {age:d} today! Happy Birthday!'
    ```

    Then, in `gramex.yaml`, you can use it as a FunctionHandler as follows::

    ```yaml
    url:
        pattern: /$YAMLURL/greet
        handler: FunctionHandler
        kwargs:
            function: greet.birthday
    ```

    Now, `/greet?name=Gramex&age=0010` returns "Gramex turns 10 today! Happy Birthday!".
    It converts the URL parameters into the types found in the annotations, e.g. `0010` into 10.

    An alternate way of configuring this is as follows::

        url:
            pattern: /$YAMLURL/greet/name/(.*)/age/(.*)
            handler: FunctionHandler
            kwargs:
                # You can pass name=... and age=... as default values
                # but ensure that handler is the first argument in the config.
                function: greet.birthday(handler, name='Name', age=10)

    Now, `/greet/name/Gramex/age/0010` returns "Gramex turns 10 today! Happy Birthday!".

    The function args and kwargs are taken from these sources this in order.

    1. From the YAML function
        - e.g. `function: greet.birthday('Name', age=10)` sets `name='Name'` and `age=10`
    2. Over-ridden by YAML URL pattern
        - e.g. `pattern: /$YAMLPATH/(.*)/(?P<age>.*)`
        - URL `/greet/Name/10` sets `name='Name'` and `age=10`
    3. Over-ridden by URL query parameters
        - e.g. URL `/greet?name=Name&age=10` sets `name='Name'` and `age=10`
    4. Over-ridden by URL POST body parameters
        - e.g. `curl -X POST /greet -d "?name=Name&age=10"` sets `name='Name'` and `age=10`

    `handler` is also available as a kwarg. You can use this as the last positional argument or
    a keyword argument. Both `def birthday(name, age, handler)` and
    `def birthday(name, age, handler=None)` are valid.
    """
    from inspect import signature
    from typing import get_type_hints
    from pandas.core.common import flatten

    params = signature(func).parameters
    hints = get_type_hints(func)

    @wraps(func)
    def wrapper(handler, *cfg_args, **cfg_kwargs):
        # If someone passes None or a non-handler object, assume empty attributes
        path_args = getattr(handler, 'path_args', [])
        path_kwargs = getattr(handler, 'path_kwargs', {})
        args = getattr(handler, 'args', {})
        request = getattr(handler, 'request', AttrDict(headers={}, body='{}'))
        # We'll create a (*args, **kwargs)
        # College args from the config `args:` and then the handler pattern /(.*)/(.*)
        req_args = list(cfg_args) + path_args
        # Collect kwargs from the config `kwargs:` and then the handler pattern /(?P<key>.*)
        # and the the URL query params
        req_kwargs = {'handler': handler}
        for d in (cfg_kwargs, path_kwargs, args):
            req_kwargs.update(d)
        headers = request.headers
        # If POSTed with Content-Type: application/json, parse body as well
        if headers.get('Content-Type', '') == 'application/json':
            req_kwargs.update(json.loads(request.body))

        # Map these into the signature
        args, kwargs = [], {}
        for arg, param in params.items():
            hint = hints.get(arg, identity)
            # If hint says it's a header, pass the header without converting
            if hint is Header:
                kwargs[arg] = request.headers.get(arg)
                continue
            # Populate positional arguments first, if any
            if len(req_args):
                # If function takes a positional arg, assign first req_arg
                if param.kind in {param.POSITIONAL_ONLY, param.POSITIONAL_OR_KEYWORD}:
                    args.append(convert(hint, param, req_args.pop(0)))
                # If it's a *arg, assign all remaining req_arg
                elif param.kind == param.VAR_POSITIONAL:
                    for val in req_args:
                        args.append(convert(hint, param, val))
                    req_args.clear()
            # Also populate keyword arguments from req_kwargs if there's a match
            if arg in req_kwargs:
                # Pop any keyword argument that matches the function argument
                if param.kind in {param.KEYWORD_ONLY, param.POSITIONAL_OR_KEYWORD}:
                    kwargs[arg] = convert(hint, param, req_kwargs.pop(arg))
                # If its a *arg, assign all remaining values
                elif param.kind == param.VAR_POSITIONAL:
                    for val in flatten([req_kwargs.pop(arg)]):
                        args.append(convert(hint, param, val))

        return func(*args, **kwargs)

    wrapper.__func__ = func
    return wrapper


# Define direct keys. These can be used as-is
_transform_direct_vars = {
    'name': 'handler.name',
    'class': 'handler.__class__.__name__',
    'time': 'round(time.time() * 1000, 0)',
    'datetime': 'datetime.datetime.utcnow().strftime("%Y-%m-%d %H:%M:%SZ")',
    'method': 'handler.request.method',
    'uri': 'handler.request.uri',
    'ip': 'handler.request.remote_ip',
    'status': 'handler.get_status()',
    'duration': 'round(handler.request.request_time() * 1000, 0)',
    'port': 'conf.app.listen.port',
    # TODO: 'size': 'handler.get_content_size()' is not available in RequestHandler
    'user': '(handler.current_user or {}).get("id", "")',
    'session': 'handler.session.get("id", "")',
    'error': 'getattr(handler, "_exception", "")',
    'browser': 'handler.request.headers.get("User-Agent", "")',
}

# Define object keys for us as key.value. E.g. cookies.sid, user.email, etc
_transform_obj_vars = {
    'args': 'handler.get_argument("{val}", "")',
    'request': 'getattr(handler.request, "{val}", "")',
    'headers': 'handler.request.headers.get("{val}", "")',
    'cookies': (
        'handler.request.cookies["{val}"].value ' + 'if "{val}" in handler.request.cookies else ""'
    ),
    'user': '(handler.current_user or {{}}).get("{val}", "")',
    'env': 'os.environ.get("{val}", "")',
}


def handler_expr(expr):
    if expr in _transform_direct_vars:
        return _transform_direct_vars[expr]
    if '.' in expr:
        prefix, value = expr.split('.', 2)
        if prefix in _transform_obj_vars:
            return _transform_obj_vars[prefix].format(val=value)
    raise ValueError(f'Unknown expression {expr}')


def build_log_info(keys: List, *vars: List):
    '''Utility to create logging values.

    Returns a function that accepts a handler and returns a dict of values to log.

    Examples:
        >>> log_info = build_log_info(['time', 'ip', 'cookies.sid', 'user.email'])
        >>> log_info(handler)
        ... {"time": 1655280440, "ip": "::1", "cookies.sid": "...", "user.email": "..."}

    Parameters:
        keys: list of keys to include in the log.
        *vars: additional variables to include in the log.

    `keys` can include: `name`, `class`, `time`, `datetime`, `method`, `uri`, `ip`, `status`,
    `duration`, port, `user`, `session`, `error`.

    It can also include:

    - `args.*`: value of the URL query parameter ?arg=
    - `request.*`: value of `handler.request.*`
    - `headers.*`: value of the HTTP header
    - `cookies.*`: value of the HTTP cookie
    - `user.*`: key in the user object
    - `env.*`: value of the environment variable

    `vars` can be any list of any variables. When you pass these to the function, they're added to
    the returned value as-is. (This is used internally in [gramex.handlers.AuthHandler.setup])

    Examples:
        >>> log_info = build_log_info(['time'], 'event')
        >>> log_info(handler, event='x')
        ... {"time": 1655280440, "event": "x"}
    '''
    from gramex import conf

    vals = []
    for key in keys:
        if key in vars:
            vals.append(f'"{key}": {key},')
            continue
        try:
            vals.append(f'"{key}": {handler_expr(key)},')
        except ValueError:
            app_log.error(f'Skipping unknown key {key}')
    code = compile(
        'def fn(handler, %s):\n\treturn {%s}' % (', '.join(vars), ' '.join(vals)),
        filename='log',
        mode='exec',
    )
    context = {'os': os, 'time': time, 'datetime': datetime, 'conf': conf, 'AttrDict': AttrDict}
    # B102:exec_used is safe here since the code is constructed entirely in this function
    exec(code, context)  # noqa S102
    return context['fn']


def time_key(format, offset=None, freq=None):
    import pandas as pd

    result = {'function': lambda *args, **kwargs: pd.Timestamp.utcnow().strftime(format)}

    if freq is not None:

        def expiry(*args, **kwargs):
            now = pd.Timestamp.utcnow()
            return int((now.ceil(freq=freq) - now).total_seconds())

    elif offset is not None:
        offset_method = getattr(pd.offsets, offset)

        def expiry(*args, **kwargs):
            now = pd.Timestamp.utcnow()
            return int((now + offset_method(normalize=True) - now).total_seconds())

    if offset is not None or freq is not None:
        result['expiry'] = expiry

    return result



import os
import posixpath
import time
import tornado.gen
import gramex.data
from string import ascii_lowercase, digits
from random import choice
from mimetypes import guess_type
from tornado.web import HTTPError
from gramex.config import objectpath, slug, variables
from gramex.http import NOT_FOUND, REQUEST_ENTITY_TOO_LARGE, UNSUPPORTED_MEDIA_TYPE
from .formhandler import FormHandler


class DriveHandler(FormHandler):
    '''
    Lets users manage files. Here's a typical configuration

    ```yaml
    path: $GRAMEXDATA/apps/appname/     # Save files here
    user_fields: [id, role, hd]         # user attributes to store
    tags: [tag]                         # <input name=""> to store
    allow: [.doc, .docx]                # Only allow these files
    ignore: [.pdf]                      # Don't allow these files
    max_file_size: 100000               # Files must be smaller than this
    url: $GRAMEXDATA/apps/appname/.meta.db  # Optional metadata DB
    table: drive                            # Optional metadata table
    redirect:                           # After uploading the file,
        query: next                     #   ... redirect to ?next=
        url: /$YAMLURL/                 #   ... else to this directory
    ```
    '''

    @classmethod
    def setup(
        cls,
        path=None,
        user_fields=None,
        tags=None,
        allow=None,
        ignore=None,
        max_file_size=None,
        url=None,
        table=None,
        storage: dict = None,
        **kwargs,
    ):
        if not path:
            folder = slug['filename'](cls.name)
            path = os.path.join(variables['GRAMEXDATA'], 'drivehandler', folder)
        cls.user_fields = cls._ensure_type('user_fields', user_fields)
        cls.tags = cls._ensure_type('tags', tags)
        cls.allow = allow or []
        cls.ignore = ignore or []
        cls.max_file_size = max_file_size or 0
        storage = storage or {'type': 'os'}
        if storage.get('type') not in storages:
            raise ValueError(f'{cls.name}: storage.type should be one of {list(storages)}')
        if storage.get('type') == 'os':
            storage['path'] = path
        cls.fs = storages[storage['type']](**storage)
        # Set up the parent FormHandler with a single URL and table
        if not url:
            url = 'sqlite:///' + os.path.join(path, '.meta.db')
            os.makedirs(path, exist_ok=True)
        table = table or 'drive'
        kwargs.update(url=url, table=table, id='id')
        cls.special_keys += ['path', 'user_fields', 'tags', 'allow', 'ignore', 'max_file_size']
        super().setup(**kwargs)

        # Ensure all tags and user_fields are present in "drive" table
        cls._db_cols = {
            'id': {'type': 'int', 'primary_key': True, 'autoincrement': True},
            'file': {'type': 'text'},  # Original file name
            'ext': {'type': 'text'},  # Original file extension
            'path': {'type': 'text'},  # Saved file relative path
            'size': {'type': 'int'},  # File size
            'mime': {'type': 'text'},  # MIME type
            'date': {'type': 'int'},  # Uploaded date
        }
        for s in cls.user_fields:
            cls._db_cols[f'user_{s}'] = {'type': 'text'}
        for s in cls.tags:
            cls._db_cols.setdefault(s, {'type': 'text'})
        gramex.data.alter(url, table, cls._db_cols)

        # If ?_download=...&id=..., then download the file via modify:
        def download_plugin(data, key, handler):
            data = original_modify(data, key, handler)
            ids = handler.args.get('id', [])
            if len(ids) != 1 or '_download' not in handler.args:
                return data
            if len(data) == 0:
                raise HTTPError(NOT_FOUND, f'No file record with id={ids[0]}')
            path = data['path'][0]
            if not cls.fs.exists(path):
                raise HTTPError(NOT_FOUND, f'Missing file for id={ids[0]}')
            handler.set_header('Content-Type', data['mime'][0])
            handler.set_header('Content-Length', cls.fs.size(path))
            handler.set_header('Content-Disposition', f'attachment; filename="{data["file"][0]}"')
            with cls.fs.open(path, 'rb') as handle:
                return handle.read()

        original_modify = cls.datasets['data'].get('modify', lambda v, *args: v)
        cls.datasets['data']['modify'] = download_plugin

    def check_filelimits(self):
        allow = {ext.lower() for ext in self.allow}
        ignore = {ext.lower() for ext in self.ignore}
        for name, ext, size in zip(self.args['file'], self.args['ext'], self.args['size']):
            if self.max_file_size and size > self.max_file_size:
                raise HTTPError(REQUEST_ENTITY_TOO_LARGE, f'{name}: {size} > {self.max_file_size}')
            if ext in ignore or (allow and ext not in allow):
                raise HTTPError(UNSUPPORTED_MEDIA_TYPE, name)

    @tornado.gen.coroutine
    def post(self, *path_args, **path_kwargs):
        '''Saves uploaded files, then updates metadata DB'''
        user = self.current_user or {}
        uploads = self.request.files.get('file', [])
        n = len(uploads)
        # Initialize all DB columns (except ID) to have the same number of rows as uploads.
        # Add `n` rows, and then clip to `n` rows. Effective way to pad AND trim.
        for key in list(self._db_cols.keys())[1:]:
            self.args[key] = self.args.get(key, []) + [''] * n
        for key in self.args:
            self.args[key] = self.args[key][:n]
        for i, upload in enumerate(uploads):
            file = os.path.basename(upload.get('filename', ''))
            ext = os.path.splitext(file)[1]
            path = slug.filename(file)
            # B311:random random() is safe since it's for non-cryptographic use
            while self.fs.exists(path):
                randomletter = choice(digits + ascii_lowercase)  # noqa S311
                path = os.path.splitext(path)[0] + randomletter + ext
            self.args['file'][i] = file
            self.args['ext'][i] = ext.lower()
            self.args['path'][i] = path
            self.args['size'][i] = len(upload['body'])
            self.args['date'][i] = int(time.time())
            # Guess MIME type from filename if it's unknown
            self.args['mime'][i] = upload['content_type']
            if self.args['mime'][i] == 'application/unknown':
                self.args['mime'][i] = guess_type(file, strict=False)[0]
            # Append user attributes
            for s in self.user_fields:
                self.args[f'user_{s.replace(".", "_")}'][i] = objectpath(user, s)
        self.check_filelimits()
        self.files = self.args
        yield super().post(*path_args, **path_kwargs)

    def pre_modify(self, **kwargs):
        '''Called by FormHandler after updating the database, before modify. Save files here.

        This allows the DriveHandler modify: action to access the files.
        '''
        # If POST or PUT, save all files
        if self.request.method in {'POST', 'PUT'}:
            uploads = self.request.files.get('file', [])
            for upload, path in zip(uploads, self.files['path']):
                with self.fs.open(path, 'wb') as handle:
                    handle.write(upload['body'])
        elif self.request.method == 'DELETE':
            for path in self.files['path']:
                if self.fs.exists(path):
                    self.fs.remove(path)

    @tornado.gen.coroutine
    def delete(self, *path_args, **path_kwargs):
        '''Deletes files from metadata DB and from file system'''
        conf = self.datasets.data
        files = gramex.data.filter(conf.url, table=conf.table, args=self.args)
        self.files = files.to_dict(orient='list')
        yield super().delete(*path_args, **path_kwargs)

    @tornado.gen.coroutine
    def put(self, *path_args, **path_kwargs):
        '''Update attributes and files'''
        uploads = self.request.files.get('file', [])[:1]
        id = self.args.get('id', [-1])
        # User cannot change the path, size, date or user attributes
        for s in ('path', 'size', 'date'):
            self.args.pop(s, None)
        for s in self.user_fields:
            self.args.pop(f'user_{s}', None)
        # These are updated only when a file is uploaded
        if len(uploads):
            user = self.current_user or {}
            self.args.setdefault('size', []).append(len(uploads[0]['body']))
            self.args.setdefault('date', []).append(int(time.time()))
            for s in self.user_fields:
                self.args.setdefault(f'user_{s.replace(".", "_")}', []).append(objectpath(user, s))
        conf = self.datasets['data']
        files = gramex.data.filter(conf.url, table=conf.table, args={'id': id})
        self.files = files.to_dict(orient='list')
        self.files.update(self.args)
        yield super().put(*path_args, **path_kwargs)

    @classmethod
    def _ensure_type(cls, field, values):
        if isinstance(values, dict):
            return values
        if isinstance(values, (list, tuple)):
            return {v: 'str' for v in values if v}
        if isinstance(values, str) and values:
            return {values: 'str'}
        if not values:
            return {}
        raise TypeError(f'{cls.name}: {field} should be a dict, not {values}')


class OSFS(object):
    def __init__(self, path, type='os'):
        self.path = path
        os.makedirs(path, exist_ok=True)

    def exists(self, path):
        return os.path.exists(os.path.join(self.path, path))

    def open(self, path, mode):
        return open(os.path.join(self.path, path), mode)

    def remove(self, path):
        return os.remove(os.path.join(self.path, path))

    def size(self, path):
        return os.stat(os.path.join(self.path, path)).st_size


class S3FS(object):
    def __init__(self, type='s3', bucket='drivehandler') -> None:
        import s3fs

        self.fs = s3fs.S3FileSystem()
        self.bucket = bucket
        self.fs.makedirs(bucket, exist_ok=True)

    def exists(self, path):
        return self.fs.exists(posixpath.join(self.bucket, path))

    def open(self, path, mode):
        return self.fs.open(posixpath.join(self.bucket, path), mode)

    def remove(self, path):
        return self.fs.rm_file(posixpath.join(self.bucket, path))

    def size(self, path):
        return self.fs.size(posixpath.join(self.bucket, path))


storages = {
    'os': OSFS,
    's3': S3FS,
}



'''Migrate from older versions of Gramex'''


def user_db():
    '''Migrate user auth database from gramex 1.86.1 to 1.87.0.

    Till gramex 1.86.1, user data was in a SQLite Dict at $GRAMEXDATA/auth.user.db with BLOB value.
    From 1.87.0, it is stored in a gramex.data.filter compatible table with TEXT value.
    This function migrates the old database to the new format.
    '''
    import gramex
    import sqlite3
    from gramex.config import objectpath, app_log, variables
    from urllib.parse import urlparse

    # No need to migrate if developer specifies a storelocation.user different from default.
    url = objectpath(gramex.service, 'storelocations.user.url', '')
    path = urlparse(url).path[1:].replace('\\', '/')
    default_path = variables['GRAMEXDATA'].replace('\\', '/').rstrip('/') + '/auth.user.db'
    if path != default_path:
        app_log.debug(f'1.87.0: SKIP migrating custom storelocations.user.url: {path}')
        return

    # No need to migrate unless user.value is a BLOB
    conn = sqlite3.connect(path)
    cur = conn.cursor()
    query = "SELECT type FROM pragma_table_info('user') WHERE name='value'"
    if cur.execute(query).fetchall()[0][0].lower() == 'text':
        app_log.debug(f'1.87.0: SKIP migrated storelocations.user.url: {path}')
        return

    # SQLite can't change column type in a single command.
    # Also, SQLite < 3.35 does not support DROP COLUMN in ALTER TABLE.
    # So create new column, copy data, drop
    app_log.info(f'1.87.0: MIGRATING user.value from BLOB to TEXT: {path}')
    cur.executescript(
        '''
        BEGIN TRANSACTION;
        CREATE TEMPORARY TABLE user_backup(key TEXT PRIMARY KEY, value TEXT);
        INSERT INTO user_backup SELECT key, value FROM user;
        DROP TABLE user;
        CREATE TABLE user(key TEXT PRIMARY KEY, value TEXT);
        INSERT INTO user SELECT key, value FROM user_backup;
        DROP TABLE user_backup;
        COMMIT;
        '''
    )

    conn.commit()
    conn.close()



'''Handlers set up the micro-services for [gramex.services.url][].'''

from .basehandler import BaseMixin, BaseHandler, BaseWebSocketHandler, SetupFailedHandler
from .functionhandler import FunctionHandler
from .websockethandler import WebSocketHandler
from .filehandler import FileHandler
from .authhandler import AuthHandler, GoogleAuth, SimpleAuth, LogoutHandler
from .processhandler import ProcessHandler
from .jsonhandler import JSONHandler
from .socialhandler import TwitterRESTHandler, FacebookGraphHandler
from .uploadhandler import UploadHandler
from .capturehandler import CaptureHandler, Capture
from .formhandler import FormHandler
from .pptxhandler import PPTXHandler
from .proxyhandler import ProxyHandler
from .modelhandler import ModelHandler
from .filterhandler import FilterHandler
from .drivehandler import DriveHandler
from .comichandler import ComicHandler
from .openapihandler import OpenAPIHandler
from .messagehandler import MessageHandler
from .chatgpthandler import ChatGPTHandler

# Aliases
ChatGPT = ChatGPTHandler
Comic = ComicHandler
Command = ProcessHandler
Message = CommentHandler = MessageHandler
Data = FormHandler
Facebook = FacebookGraphHandler
File = DirectoryHandler = FileHandler
Filter = FilterHandler
Function = FunctionHandler
JSON = JSONHandler
OpenAPI = OpenAPIHandler
Proxy = ProxyHandler
Screenshot = CaptureHandler
Slide = PPTXHandler
Storage = DriveHandler
Twitter = TwitterRESTHandler
Upload = UploadHandler
Websocket = WebSocketHandler


__all__ = [
    'AuthHandler',
    'BaseHandler',
    'BaseWebSocketHandler',
    'BaseMixin',
    'Capture',
    'CaptureHandler',
    'ChatGPT',
    'ChatGPTHandler',
    'Command',
    'CommentHandler',
    'Comic',
    'ComicHandler',
    'Data',
    'DirectoryHandler',
    'DriveHandler',
    'FacebookGraphHandler',
    'File',
    'FileHandler',
    'Filter',
    'FilterHandler',
    'FormHandler',
    'Function',
    'FunctionHandler',
    'GoogleAuth',
    'JSON',
    'JSONHandler',
    'LogoutHandler',
    'Message',
    'MessageHandler',
    'ModelHandler',
    'OpenAPI',
    'OpenAPIHandler',
    'PPTXHandler',
    'ProcessHandler',
    'Proxy',
    'ProxyHandler',
    'Screenshot',
    'SetupFailedHandler',
    'SimpleAuth',
    'Slide',
    'Storage',
    'Twitter',
    'TwitterRESTHandler',
    'Upload',
    'UploadHandler',
    'Websocket',
    'WebSocketHandler',
]


# MLHandler requires optional dependencies scikit-learn and statsmodels.
# If they're not installed, ignore the import error and skip MLHandler.
try:
    from .mlhandler import MLHandler, MLPredictor

    ML = MLHandler
    __all__ += ['ML', 'MLHandler', 'MLPredictor']
except ImportError as e:
    from gramex.config import app_log

    app_log.warning('url: MLHandler/MLPredictor dependency missing. %s', e)

try:
    # If Gramex enterprise is available, import all handlers
    import gramexenterprise.handlers

    if hasattr(gramexenterprise, 'handlers'):
        from gramexenterprise.handlers import *  # noqa

        __all__ += gramexenterprise.handlers.__all__
except ImportError:
    pass



import os
import smtplib
from email import encoders
from mimetypes import guess_type
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.mime.base import MIMEBase
from email.mime.image import MIMEImage
from email.utils import formataddr, getaddresses
from gramex.config import app_log
from typing import List, Union


class SMTPMailer:
    '''Creates an object capable of sending HTML emails.

    Examples:
        >>> mailer = SMTPMailer(type='gmail', email='gramex.guide@gmail.com', password='...')
        >>> mailer.mail(
        ... to='person@example.com',
        ... subject='Subject',
        ... html='<strong>Bold text</strong>. <img src="cid:logo">'
        ... body='This plain text is shown if the client cannot render HTML',
        ... attachments=['1.pdf', '2.txt'],
        ... images={'logo': '/path/to/logo.png'})
    '''

    clients = {
        'gmail': {'host': 'smtp.gmail.com'},
        'yahoo': {'host': 'smtp.mail.yahoo.com'},
        'live': {'host': 'smtp.live.com'},
        'mandrill': {'host': 'smtp.mandrillapp.com'},
        'office365': {'host': 'smtp.office365.com'},
        'outlook': {'host': 'smtp-mail.outlook.com'},
        'icloud': {'host': 'smtp.mail.me.com'},
        'mail.com': {'host': 'smtp.mail.com'},
        'smtp': {'tls': False},
        'smtps': {},
    }
    # SMTP port, depending on whether TLS is True or False
    ports = {True: 587, False: 25}

    def __init__(
        self,
        type: str,
        email: str = None,
        password: str = None,
        host: str = None,
        port: int = None,
        tls: bool = True,
        stub: str = None,
    ):
        '''
        Parameters:
            type: Email service type
            email: SMTP server login email ID
            password: SMTP server login email password
            host: SMTP server. Not required when a `type` is specified
            port: SMTP server port. Defaults to 25 for non-TLS, 587 for TLS
            tls: True to use TLS, False to use non-TLS
            stub: 'log' prints email contents instead of sending it

        `type` can be:

        - `gmail`: Google Mail (smtp.gmail.com)
        - `yahoo`: Yahoo Mail (smtp.mail.yahoo.com)
        - `live`: Live.com mail (smtp.live.com)
        - `mandrill`: Mandrill (smtp.mandrillapp.com)
        - `office365`: Office 365 (smtp.office365.com)
        - `outlook`: Outlook (smtp-mail.outlook.com)
        - `icloud`: Apple iCloud (smtp.mail.me.com)
        - `mail`.com: Mail.com (smtp.mail.com)
        - `smtp`: Use ANY SMTP server via `host=`. tls=False by default
        - `smtps`: Use ANY SMTP server via `host=`. tls=True by default

        To test emails without sending them, use `stub=True` option. This queues email info into
        `SMTPStub.stubs` without sending it. Use `stub='log'` to print the email contents as well.
        '''
        self.type = type
        self.email = email
        self.password = password
        self.stub = stub
        if type not in self.clients:
            raise ValueError(f'Unknown email type: {type}')
        self.client = self.clients[type]
        for key, val in (('host', host), ('port', port), ('tls', tls)):
            if val is not None:
                self.client[key] = val
        if 'host' not in self.client:
            raise ValueError('Missing SMTP host')

    def mail(self, **kwargs):
        '''Sends an email.

        Examples:
            >>> mailer = SMTPMailer(type='gmail', email='gramex.guide@gmail.com', password='...')
            >>> mailer.mail(
            ... to='person@example.com',
            ... subject='Subject',
            ... html='<strong>Bold text</strong>. <img src="cid:logo">'
            ... body='This plain text is shown if the client cannot render HTML',
            ... attachments=['1.pdf', '2.txt'],
            ... images={'logo': '/path/to/logo.png'})
            >>> message(to='b@example.org', subject=sub, body=text, html=html)
            >>> message(to='b@example.org', subject=sub, body=text, attachments=['file.pdf'])
            >>> message(to='b@example.org', subject=sub, body=text, attachments=[
                    {'filename': 'test.txt', 'body': 'File contents'}
                ])
            >>> message(to='b@example.org', subject=sub, html='<img src="cid:logo">',
                        images={'logo': 'd:/images/logo.png'})

        Parameters may be any email parameter in [RFC 2822](https://www.rfc-wiki.org/wiki/RFC2822).
        Parameters are case insensitive. Most commonly used are:

        - `to`: The recipient email address
        - `cc`: The carbon copy recipient email address
        - `bcc`: The blind carbon copy recipient email address
        - `reply_to`: The reply-to email address
        - `on_behalf_of`: The sender email address
        - `subject`: The email subject
        - `body`: text content of the email
        - `html`: HTML content of the email. If both `html` and `body`
            are specified, the email contains both parts. Email clients may decide to
            show one or the other.
        - `attachments`: an list of file names or dict with:
            - `body`: a byte array of the content
            - `content_type`: MIME type or `filename` indicating the file name
        - `images`: dict of `{key: path}`.
            - `key` may be anything. The HTML should include the image via `<img src="cid:key">`
            - `path` is the absolute path to the image

        In addition, any keyword arguments passed are treated as message headers.

        `To`, `Cc` and `Bcc` may be:

        - a string with comma-separated emails, e.g. `'a@x.com, b@x.com'`
        - a list of strings with emails, e.g. `['a@x.com', 'b@x.com']`
        - a list of strings with comma-separated emails, e.g. `['a@x.com', 'b@x.com, c@x.com']`
        '''
        sender = kwargs.get('sender', self.email)
        # SES allows restricting the From: address. https://amzn.to/2Kqwh2y
        # SES uses an IAM ID for login (self.email), but restricts sender -- so from= is required
        # Mailgun suggests From: be the same as Sender: http://bit.ly/2tGS5wt
        # If mail(from=) is specified, use that as the sender. Else use email: from the service
        if kwargs.get('from', None):
            sender = kwargs['from']
        else:
            kwargs['from'] = sender
        # Identify recipients from to/cc/bcc fields.
        # Note: We MUST explicitly add EVERY recipient (to/cc/bcc) in sendmail(recipients=)
        to = recipients(**kwargs)
        msg = message(**kwargs)
        tls = self.client.get('tls', True)
        # Test cases specify stub: true. This uses a stub that logs emails
        if self.stub:
            server = SMTPStub(
                self.client['host'], self.client.get('port', self.ports[tls]), self.stub
            )
        else:
            server = smtplib.SMTP(self.client['host'], self.client.get('port', self.ports[tls]))
        if tls:
            server.starttls()
        if self.email is not None and self.password is not None:
            server.login(self.email, self.password)
        server.sendmail(sender, to, msg.as_string())
        server.quit()
        app_log.info(f'Email sent via {self.client["host"]} ({self.email}) to {", ".join(to)}')

    def mail_log(self, **kwargs):
        '''Same as mail() but logs the exception. Useful with running in a thread'''
        try:
            self.mail(**kwargs)
        except Exception:
            app_log.exception(f'SMTP failed: {kwargs}')
            raise


def recipients(**kwargs):
    # Return all recipients from to/cc/bcc fields.
    # They may be comma-separated strings or lists of comma-separated strings.
    recipients = []
    for key in kwargs:
        if key.lower() in {'to', 'cc', 'bcc'}:
            to = kwargs[key]
            if isinstance(to, str):
                to = [to]
            # Format as RFC-822 formatted email addresses
            recipients += [formataddr(pair) for pair in getaddresses(to)]
    return recipients


def message(
    body: str = None,
    html: str = None,
    attachments: List[Union[str, dict]] = [],
    images: dict = {},
    **kwargs: dict,
):
    # Returns a MIME message object based on text or HTML content, and optional attachments.
    if html:
        if not images:
            msg = html_part = MIMEText(html.encode('utf-8'), 'html', 'utf-8')
        else:
            msg = html_part = MIMEMultipart('related')
            html_part.attach(MIMEText(html.encode('utf-8'), 'html', 'utf-8'))
            for name, path in images.items():
                with open(path, 'rb') as handle:
                    img = MIMEImage(handle.read())
                    img.add_header('Content-ID', f'<{name}>')
                    html_part.attach(img)
    if body and html:
        msg = MIMEMultipart('alternative')
        msg.attach(MIMEText(body.encode('utf-8'), 'plain', 'utf-8'))
        msg.attach(html_part)
    elif not html:
        msg = MIMEText((body or '').encode('utf-8'), 'plain', 'utf-8')

    if attachments:
        msg_addon = MIMEMultipart()
        msg_addon.attach(msg)
        for doc in attachments:
            if isinstance(doc, dict):
                filename = doc.get('filename', 'data.bin')
                content_type = doc.get('content_type', guess_type(filename, strict=False)[0])
                content = doc['body']
            else:
                filename = doc
                with open(filename, 'rb') as handle:
                    content = handle.read()
                content_type = guess_type(filename, strict=False)[0]
            if content_type is None:
                content_type = 'application/octet-stream'
            maintype, subtype = content_type.split('/', 1)
            msg = MIMEBase(maintype, subtype)
            msg.set_payload(content)
            encoders.encode_base64(msg)
            msg.add_header(
                'Content-Disposition', 'attachment', filename=os.path.basename(filename)
            )
            msg_addon.attach(msg)
        msg = msg_addon

    # set headers
    for arg, value in kwargs.items():
        header = '-'.join(
            [
                # All SMTP headers are capitalised, except abbreviations
                w.upper() if w in {'ID', 'MTS', 'IPMS'} else w.capitalize()
                for w in arg.split('_')
            ]
        )
        msg[header] = _merge(value)

    return msg


def _merge(value):
    return ', '.join(value) if isinstance(value, list) else value


class SMTPStub:
    # A minimal test stub for smtplib.SMTP with features used in this module
    stubs = []

    def __init__(self, host, port, options):
        # Maintain a list of all stub info so far
        self.options = options
        self.info = {}
        self.stubs.append(self.info)
        self.info.update(host=host, port=port)

    def starttls(self):
        self.info.update(starttls=True)

    def login(self, email, password):
        self.info.update(email=email, password=password)

    def sendmail(self, from_addr, to_addrs, msg):
        self.info.update(from_addr=from_addr, to_addrs=to_addrs, msg=msg)

    def quit(self):
        self.info.update(quit=True)
        if self.options == 'log':
            app_log.debug(f'From: {self.info["from_addr"]}')
            app_log.debug(f'To: {self.info["to_addrs"]}')
            app_log.debug(self.info['msg'])



import os
import gramex.cache
import pandas as pd
from nose.tools import eq_
from . import TestGramex, folder, afe


def eqframe(result, expected, **kwargs):
    '''Same as assert_frame_equal or afe, but does not compare index'''
    actual = pd.DataFrame(result)
    expected.index = actual.index
    afe(actual, expected, **kwargs)


def unique_of(data: pd.DataFrame, cols):
    return data.groupby(cols).size().reset_index().drop(0, axis=1)


class TestFilterHandler(TestGramex):
    sales_file = os.path.join(folder, 'sales.xlsx')
    sales = gramex.cache.open(sales_file, 'xlsx')
    census = gramex.cache.open(sales_file, 'xlsx', sheet_name='census')

    def test_filters(self):
        self.check('/filters/sales?_c=city')

        result = self.get('/filters/sales', params={'_c': ['city']}).json()
        expected = unique_of(self.sales, 'city')
        eqframe(result['city'], expected)

        result = self.get('/filters/sales', params={'_c': ['city', 'product']}).json()
        for col in ['city', 'product']:
            expected_col = unique_of(self.sales, col)
            eqframe(result[col], expected_col)

        limit = 10
        _args = {'_c': ['city'], '_limit': [limit]}
        result = self.get('/filters/sales', params=_args).json()
        expected = unique_of(self.sales, 'city').head(limit)
        eqframe(result['city'], expected)

        _args = {'_c': ['city'], '_sort': ['city']}
        result = self.get('/filters/sales', params=_args).json()
        expected = unique_of(self.sales, 'city').sort_values('city')
        eqframe(result['city'], expected)

        _args = {'_c': ['city'], '_sort': ['city', '-sales']}
        result = self.get('/filters/sales', params=_args).json()
        expected = unique_of(expected, 'city')
        expected = expected.sort_values(['city'])  # sales is ignored. It sorts AFTER grouping
        eqframe(result['city'], expected)

        _args = {'_c': ['District'], 'State': ['KERALA']}
        result = self.get('/filters/census', params=_args).json()
        expected = self.census[self.census['State'] == 'KERALA']
        expected = unique_of(expected, 'District')
        eqframe(result['District'], expected)

        _args = {
            '_c': ['District', 'DistrictCaps'],
            '_sort': ['State', 'District'],
            'State': ['KERALA'],
            '_limit': [10],
        }
        result = self.get('/filters/census', params=_args).json()
        filtered = self.census[self.census['State'] == 'KERALA']
        for col, sort_cols in [('District', ['District']), ('DistrictCaps', [])]:
            eqframe(result[col], unique_of(filtered, col).sort_values(sort_cols).head(10))

        for col in ['city', 'product']:
            _args = {'_c': [col], 'देश': ['भारत']}
            result = self.get('/filters/sales', params=_args).json()
            expected = self.sales[self.sales['देश'] == 'भारत']
            expected = unique_of(expected, col)
            eqframe(result[col], expected)

    def test_ranges(self):
        sales, growth = self.sales.sales, self.sales.growth
        result = self.get('/filters/sales', params={'_c': ['sales|Range']}).json()
        expected = {'sales|Range': [{'sales|min': sales.min(), 'sales|max': sales.max()}]}
        eq_(result, expected)

        result = self.get('/filters/sales', params={'_c': ['sales|Min']}).json()
        expected = {'sales|Min': [{'sales|Min': sales.min()}]}
        eq_(result, expected)

        result = self.get('/filters/sales', params={'_c': ['sales,growth|Max']}).json()
        expected = {'sales,growth|Max': [{'sales|Max': sales.max(), 'growth|Max': growth.max()}]}
        eq_(result, expected)

        result = self.get('/filters/sales', params={'_c': ['sales,growth|Range']}).json()
        expected = {
            'sales,growth|Range': [
                {
                    'sales|min': sales.min(),
                    'sales|max': sales.max(),
                    'growth|min': growth.min(),
                    'growth|max': growth.max(),
                }
            ]
        }
        eq_(result, expected)

    def test_multifilters(self):
        _args = {'_c': ['देश,city']}
        result = self.get('/filters/sales', params=_args).json()
        expected = unique_of(self.sales, ['देश', 'city'])
        eqframe(result['देश,city'], expected)

        _args = {'_c': ['city,city']}
        result = self.get('/filters/sales', params=_args).json()
        expected = unique_of(self.sales, ['city'])
        eqframe(result['city,city'], expected)

        _args = {'_c': ['देश,city,product']}
        result = self.get('/filters/sales', params=_args).json()
        expected = unique_of(self.sales, ['देश', 'city', 'product'])
        eqframe(result['देश,city,product'], expected)

        _args = {'_c': ['देश,city', 'product']}
        result = self.get('/filters/sales', params=_args).json()
        expected = unique_of(self.sales, ['देश', 'city'])
        eqframe(result['देश,city'], expected)
        expected = unique_of(self.sales, ['product'])
        eqframe(result['product'], expected)

        limit = 10
        _args = {'_c': ['देश,city'], '_limit': limit}
        result = self.get('/filters/sales', params=_args).json()
        expected = unique_of(self.sales, ['देश', 'city']).head(limit)
        eqframe(result['देश,city'], expected)

        _args = {'_c': ['देश,city'], '_sort': ['देश', '-city']}
        result = self.get('/filters/sales', params=_args).json()
        expected = unique_of(self.sales, ['देश', 'city']).sort_values(
            ['देश', 'city'], ascending=[True, False]
        )
        eqframe(result['देश,city'], expected)

        _args = {'_c': ['देश,city'], '_sort': ['देश', '-sales']}
        result = self.get('/filters/sales', params=_args).json()
        expected = unique_of(self.sales, ['देश', 'city'])
        expected = expected.sort_values(['देश'])
        eqframe(result['देश,city'], expected)

        _args = {'_c': ['देश,city'], 'sales>': ['500']}
        result = self.get('/filters/sales', params=_args).json()
        expected = self.sales[self.sales['sales'] > 500]
        expected = unique_of(expected, ['देश', 'city'])
        eqframe(result['देश,city'], expected)

        _args = {'_c': ['देश,city'], 'sales>': ['500'], 'growth>': ['0'], '_sort': ['city']}
        result = self.get('/filters/sales', params=_args).json()
        expected = self.sales[(self.sales['sales'] > 500) & (self.sales['growth'] > 0)]
        expected = unique_of(expected, ['देश', 'city']).sort_values(['city'])
        eqframe(result['देश,city'], expected)



import re
import sys
import os.path
from glob import glob

# B410:import_lxml lxml.etree is safe on https://github.com/tiran/defusedxml/tree/main/xmltestdata
from lxml.etree import Element  # noqa S410
from lxml.html import fromstring, tostring  # noqa S410
import numpy as np
import pandas as pd
import gramex.data
import gramex.cache
from gramex import conf
from gramex.config import app_log
from gramex.transforms import build_transform
from typing import List

if sys.version_info.major == 3:
    unicode = str

DB_CONFIG = {
    'table': 'agg{}',
    'levels': ['M', 'W', 'D'],
    'dimensions': [{'key': 'time', 'freq': '?level'}, 'user.id', 'ip', 'status', 'uri'],
    'metrics': {'duration': ['count', 'sum'], 'new_session': ['sum'], 'session_time': ['sum']},
}

# TODO: extra_columns should not be a global. Once instance may use multiple logviewers!
extra_columns = []
for key in conf.get('schedule', []):
    if 'kwargs' in conf.schedule[key] and 'custom_dims' in conf.schedule[key].kwargs:
        extra_columns = list(conf.schedule[key].kwargs.custom_dims)

DB_CONFIG['dimensions'].extend(extra_columns)

DB_CONFIG['table_columns'] = [f'{k}_{x}' for k, v in DB_CONFIG['metrics'].items() for x in v] + [
    x['key'] if isinstance(x, dict) else x for x in DB_CONFIG['dimensions']
]


FOLDER = os.path.dirname(os.path.abspath(__file__))
CONFIG_FILE = os.path.join(FOLDER, 'config.yaml')


def pdagg(df, groups, aggfuncs):
    '''
    groups = [{'key': 'time', 'freq': 'D'}, 'user.id', 'status', 'uri']
    aggfuncs = {'duration': ['count', 'mean', namedfunc], 'status': ['count']}
    '''
    groups = [pd.Grouper(**g) if isinstance(g, dict) else g for g in groups]
    grps = df.groupby(groups)
    dff = grps.agg(aggfuncs)
    if isinstance(dff.columns, pd.MultiIndex):
        dff.columns = dff.columns.map('_'.join)
    return dff.reset_index()


def add_session(df, duration=30, cutoff_buffer=0):
    '''add new_session based on `duration` threshold
    add cutoff_buffer in minutes for first and last session requests
    '''
    s = df.groupby('user.id')['time'].diff().dt.total_seconds()
    flag = s.isnull() | s.ge(duration * 60)
    df['new_session'] = flag.astype(int)
    df['session_time'] = np.where(flag, cutoff_buffer * 60, s)
    return df


def prepare_logs(df, session_threshold=15, cutoff_buffer=0, custom_dims={}):
    '''
    - removes rows with errors in time, duration, status
    - sort by time
    - adds session metrics (new_session, session_time)
    '''
    df['time'] = pd.to_datetime(df['time'], unit='ms', errors='coerce')
    # Ignore pre-2000 year and null/NaT rows
    df = df[df['time'] > '2000-01-01']
    for col in ['duration', 'status']:
        if not np.issubdtype(df[col].dtype, np.number):
            df[col] = pd.to_numeric(df[col], errors='coerce')
            df = df[df[col].notnull()]
    # logging via threads may not maintain order
    df = df.sort_values(by='time')

    for key, value in custom_dims.items():
        fn = build_transform({'function': value}, vars={'df': None}, iter=False)
        df[key] = fn(df)

    # add new_session
    return add_session(df, duration=session_threshold, cutoff_buffer=cutoff_buffer)


def summarize(
    db: dict,
    transforms: List[dict] = [],
    post_transforms: List[dict] = [],
    session_threshold: float = 15,
    cutoff_buffer: float = 0,
    custom_dims: dict = None,
) -> None:
    '''Summarizes log files into a database periodically.

    Parameters:
        db: SQLAlchemy database configuration.
        transforms: List of transforms to be applied on data.
        post_transforms: List of post transforms to be applied on data.
        session_threshold: Minimum threshold for the session.
        cutoff_buffer: In minutes for first and last session requests.
        custom_dims: Custom columns to be added to the logviewer.

    This function is called by a scheduler and/or on start of gramex.
    It will aggregate and update logs from requests.csv file by comparing the
    timestamp of last added logs. It creates the aggregation tables if they don't exist.
    '''
    app_log.info('logviewer.summarize started')
    levels = DB_CONFIG['levels']
    table = DB_CONFIG['table'].format
    # dimensions and metrics to summarize
    groups = DB_CONFIG['dimensions']
    aggfuncs = DB_CONFIG['metrics']
    log_file = conf.log.handlers.requests.filename
    # Handle for multiple instances requests.csv$LISTENPORT
    log_file = '{0}{1}'.format(*log_file.partition('.csv'))

    # all log files sorted by modified time
    log_files = sorted(glob(log_file + '*'), key=os.path.getmtime)

    def filesince(filename, date):
        match = re.search(r'(\d{4}-\d{2}-\d{2})$', filename)
        backupdate = match.group() if match else ''
        return backupdate >= date or backupdate == ''

    # get most recent log files if db is already created
    try:
        log_filter = gramex.data.filter(**db, table=table(levels[-1]), args={})
        max_date = log_filter.sort_values('time', ascending=False)['time'].iloc[0]
        max_date = pd.to_datetime(max_date)
    except Exception:
        max_date = None
    else:
        app_log.info(f'logviewer.summarize: processing since {max_date}')
        this_month = max_date.strftime('%Y-%m-01')
        log_files = [f for f in log_files if filesince(f, this_month)]

    if not log_files:
        app_log.info('logviewer.summarize: no log files to process')
        return
    # Create dataframe from log files
    columns = conf.log.handlers.requests['keys']
    app_log.info(f'logviewer.summarize: processing {log_files}')
    data = pd.concat(
        [
            pd.read_csv(f, names=columns, index_col=False, encoding='utf-8').fillna('-')
            for f in log_files
        ],
        ignore_index=True,
    )
    app_log.info(
        'logviewer.summarize: prepare_logs {} rows with session_threshold={}'.format(
            len(data.index), session_threshold
        )
    )
    data = prepare_logs(
        df=data,
        session_threshold=session_threshold,
        cutoff_buffer=cutoff_buffer,
        custom_dims=custom_dims,
    )
    app_log.info('logviewer.summarize: processed {} rows'.format(len(data.index)))
    # apply transforms on raw data
    app_log.info('logviewer.summarize: applying transforms')
    for spec in transforms:
        apply_transform(data, spec)  # applies on copy
    # levels should go from M > W > D
    for freq in levels:
        app_log.info('logviewer.summarize: aggregating {}'.format(table(freq)))
        # filter dataframe for max_date.level
        if max_date:
            date_from = max_date
            if freq == 'W':
                date_from -= pd.offsets.Day(max_date.weekday())
            if freq == 'M':
                date_from -= pd.offsets.MonthBegin(1)
            data = data[data.time.ge(date_from)]
            # delete old records
            gramex.data.delete(**db, table=table(freq), args={'time>~': [date_from]}, id=['time'])
        groups[0]['freq'] = freq
        # get summary view
        dff = pdagg(data, groups, aggfuncs)
        # apply post_transforms here
        for spec in post_transforms:
            apply_transform(dff, spec)
        # insert new records
        cols = {}
        for col in dff.columns:
            dt = dff[col].dtype.type
            if pd.api.types.is_datetime64_any_dtype(dt):
                cols[col] = 'DATETIME'
            elif pd.api.types.is_bool_dtype(dt):
                cols[col] = 'BOOLEAN'
            elif pd.api.types.is_integer_dtype(dt):
                cols[col] = 'INTEGER'
            elif pd.api.types.is_numeric_dtype(dt):
                cols[col] = 'REAL'
            else:
                cols[col] = 'TEXT'
        gramex.data.alter(**db, table=table(freq), columns=cols)
        gramex.data.insert(**db, table=table(freq), args=dff.to_dict())
    app_log.info('logviewer.summarize: completed')
    return


def prepare_where(query, args, columns):
    '''prepare where clause'''
    wheres = []
    for key, vals in args.items():
        col, agg, op = gramex.data._filter_col(key, columns)
        if col not in columns:
            continue
        if op == '':
            wheres.append('"{}" IN ("{}")'.format(col, '", "'.join(vals)))
        elif op == '!':
            wheres.append('"{}" NOT IN ("{}")'.format(col, '", "'.join(vals)))
        elif op == '>':
            wheres.append('"{}" > "{}"'.format(col, min(vals)))
        elif op == '>~':
            wheres.append('"{}" >= "{}"'.format(col, min(vals)))
        elif op == '<':
            wheres.append('"{}" < "{}"'.format(col, max(vals)))
        elif op == '<~':
            wheres.append('"{}" <= "{}"'.format(col, max(vals)))
        elif op == '~':
            q = ' OR '.join('"{}" LIKE "%{}%"'.format(col, x) for x in vals)
            wheres.append('({})'.format(q))
        elif op == '!~':
            q = ' OR '.join('"{}" NOT LIKE "%{}%"'.format(col, x) for x in vals)
            wheres.append('({})'.format(q))
        elif op == '*':
            q = ' OR '.join('"{}" ILIKE "%{}%"'.format(col, x) for x in vals)
            wheres.append('({})'.format(q))
        elif op == '!*':
            q = ' OR '.join('"{}" NOT ILIKE "%{}%"'.format(col, x) for x in vals)
            wheres.append('({})'.format(q))
    wheres = ' AND '.join(wheres)
    if not wheres:
        return wheres
    prepend = 'WHERE ' if ' WHERE ' not in query else 'AND '
    return prepend + wheres


def query(handler, args):
    '''queries for logviewer'''
    queries = handler.kwargs.kwargs.queries
    table = handler.path_kwargs.get('table')
    case = handler.path_kwargs.get('query')
    query = queries.get(case)
    wheres = prepare_where(query, args, DB_CONFIG['table_columns'])
    return query.format(table=table, where=wheres)


def apply_transform(data, spec):
    '''apply transform on dataframe'''
    pandas_transforms = {
        'REPLACE': pd.Series.replace,
        'MAP': pd.Series.map,
        'IN': pd.Series.isin,
        'NOTIN': lambda s, v: ~s.isin(v),
        'CONTAINS': {
            'function': lambda s, v, **ops: s.str.contains(v, **ops),
            'defaults': {'case': False},
        },
        'NOTCONTAINS': {
            'function': lambda s, v, **ops: ~s.str.contains(v, **ops),
            'defaults': {'case': False},
        },
        'LEN': lambda s, _: s.str.len(),
        'LOWER': lambda s, _: s.str.lower(),
        'UPPER': lambda s, _: s.str.upper(),
        'PROPER': lambda s, _: s.str.capitalize(),
        'STARTSWITH': lambda s, v: s.str.startswith(v),
        'ENDSWITH': lambda s, v: s.str.endswith(v),
    }
    # TODO: STRREPLACE
    if spec['type'] == 'function':
        fn = build_transform(
            {'function': spec['expr']}, vars={'data': None}, filename=f'lv: {spec.get("name")}'
        )
        fn(data)  # applies on copy
        return data
    expr = spec['expr']
    func = pandas_transforms[expr['op']]
    kwargs = expr.get('kwargs', {})
    if isinstance(func, dict):
        # use defaults' kwargs if not present in expr.get
        for key, val in func.get('defaults', {}).items():
            if key not in kwargs:
                kwargs[key] = val
        func = func['function']
    data[spec['as']] = func(data[expr['col']], expr.get('value'), **kwargs)
    return data


def get_config(handler=None):
    '''return config as dict'''
    file_path = handler.kwargs.get('path_ui', CONFIG_FILE)
    return gramex.cache.open(file_path, 'config')


def load_component(page, **kwargs):
    '''return generateed template'''
    return gramex.cache.open(page, 'template', rel=True).generate(**kwargs)


def load_layout(config):
    '''return generated layout'''
    return tostring(eltree(config, root=Element('root')))[6:-7]


def eltree(data, root=None):
    '''Convert dict to etree.Element(s)'''
    attr_prefix = '@'
    text_key = '$'
    tpl_key = '_$'
    result = [] if root is None else root
    if isinstance(data, dict):
        for key, value in data.items():
            if root is not None:
                # attribute prefixes
                if key.startswith(attr_prefix):
                    key = key.lstrip(attr_prefix)
                    result.set(key, unicode(value))
                    continue
                # text content
                if key == text_key:
                    result.text = unicode(value)
                    continue
                # template hooks
                if key == tpl_key:
                    for tpl in value if isinstance(value, list) else [value]:
                        template = '{}.html'.format(tpl['tpl'])
                        raw_node = load_component(template, values=tpl.get('values', tpl))
                        result.append(fromstring(raw_node))
                    continue
            # add other keys as children
            values = value if isinstance(value, list) else [value]
            for value in values:
                elem = Element(key)
                result.append(elem)
                # scalars to text
                if not isinstance(value, (dict, list)):
                    value = {text_key: value}
                eltree(value, root=elem)
    else:
        result.append(Element(unicode(data)))
    return result



import io
import os
import tornado.web
import tornado.gen
from threading import RLock
from typing import Union, List
from .basehandler import BaseHandler
from gramex.config import app_log
from gramex.cache import Subprocess


class ProcessHandler(BaseHandler):
    @classmethod
    def setup(
        cls,
        args: Union[List[str], str],
        shell: bool = False,
        cwd: str = None,
        buffer: Union[str, int] = 0,
        headers: dict = {},
        **kwargs,
    ):
        '''Set up handler to stream process output.

        Parameters:

            args: The first value is the command. The rest are optional string arguments.
                Same as `subprocess.Popen()`.
            shell: `True` passes the `args` through the shell, allowing wildcards like `*`.
                If `shell=True` then use a single string for `args` that includes the arguments.
            cwd: Current working directory from where the command will run.
                Defaults to the same directory Gramex ran from.
            buffer: Number of bytes to buffer. Defaults: `io.DEFAULT_BUFFER_SIZE`. Or `"line"`
                to buffer by newline.
            headers: HTTP headers to set on the response.
        '''
        super(ProcessHandler, cls).setup(**kwargs)
        cls.cmdargs = args
        cls.shell = shell
        cls._write_lock = RLock()
        cls.buffer_size = buffer
        # Normalize current directory for path, if provided
        cls.cwd = cwd if cwd is None else os.path.abspath(cwd)
        # File handles for stdout/stderr are cached in cls.handles
        cls.handles = {}

        cls.headers = headers
        cls.post = cls.get

    def stream_callbacks(self, targets, name):
        # stdout/stderr are can be specified as a scalar or a list.
        # Convert it into a list of callback fn(data)

        # if no target is specified, stream to RequestHandler
        if targets is None:
            targets = ['pipe']
        # if a string is specified, treat it as the sole file output
        elif not isinstance(targets, list):
            targets = [targets]

        callbacks = []
        for target in targets:
            # pipe write to the RequestHandler
            if target == 'pipe':
                callbacks.append(self._write)
            # false-y values are ignored. (False, 0, etc)
            elif not target:
                pass
            # strings are treated as files
            elif isinstance(target, (str, bytes)):
                # cache file handles for re-use between stdout, stderr
                if target not in self.handles:
                    self.handles[target] = io.open(target, mode='wb')
                handle = self.handles[target]
                callbacks.append(handle.write)
            # warn on unknown parameters (e.g. numbers, True, etc)
            else:
                app_log.warning(f'ProcessHandler: {name}: {target} is not implemented')
        return callbacks

    def initialize(
        self,
        stdout: Union[List[str], str] = None,
        stderr: Union[List[str], str] = None,
        stdin: Union[List[str], str] = None,
        **kwargs,
    ):
        '''Sets up I/O stream processing.

        Parameters:

            stdout: The process output can be sent to
            stderr: The process error stream has the same options as stdout.
            stdin: (**TODO**)

        `stdout`, `stderr` and `stdin` can be one of the below, or a list of the below:

        - `"pipe"`: Display the (transformed) output. This is the default
        - `"false"`: Ignore the output
        - `"filename.txt"`: Save output to a `filename.txt`
        '''
        super(ProcessHandler, self).initialize(**kwargs)
        self.stream_stdout = self.stream_callbacks(stdout, name='stdout')
        self.stream_stderr = self.stream_callbacks(stderr, name='stderr')

    @tornado.gen.coroutine
    def get(self, *path_args):
        if self.redirects:
            self.save_redirect_page()
        for header_name, header_value in self.headers.items():
            self.set_header(header_name, header_value)

        proc = Subprocess(
            self.cmdargs,
            # NOTE: developer should sanitize args if shell=True
            # B604 any_other_function_with_shell_equals_true
            shell=self.shell,  # noqa S604
            cwd=self.cwd,
            stream_stdout=self.stream_stdout,
            stream_stderr=self.stream_stderr,
            buffer_size=self.buffer_size,
        )
        yield proc.wait_for_exit()
        # Wait for process to finish
        proc.proc.wait()
        if self.redirects:
            self.redirect_next()

    def _write(self, data):
        with self._write_lock:
            self.write(data)
            # Flush every time. This disables Etag, but processes have
            # side-effects, so we should not be caching these requests anyway.
            self.flush()

    def on_finish(self):
        '''Close all open handles after the request has finished'''
        for handle in self.handles.values():
            handle.close()
        super(ProcessHandler, self).on_finish()



import os
import json
import time
import datetime
import tornado.httpclient
import tornado.ioloop
import gramex
from oauthlib import oauth1
from tornado.httputil import HTTPHeaders, parse_response_start_line
from urllib.parse import urlencode
from gramex.config import app_log
from gramex.transforms import flattener, build_transform
from gramex.http import (
    RATE_LIMITED,
    TOO_MANY_REQUESTS,
    CLIENT_TIMEOUT,
    INTERNAL_SERVER_ERROR,
    GATEWAY_TIMEOUT,
)


class TwitterStream:
    '''
    Starts a Twitter Streaming client.

    Examples:
        >>> from gramex.transforms.twitterstream import TwitterStream
        >>> stream = TwitterStream(
        ...     track='modi,mms',
        ...     path='save-as-file.json',
        ...     key='...',
        ...     secret='...',
        ...     access_key='...',
        ...     access_secret='...',
        ...     flush=True)

    This saves all tweets mentioning `modi` or `mms` in `save-as-file.json`
    with each line representing a tweet in JSN format.

    If `flush=True`, the file is flushed on every tweet. If `flush=<number>`,
    the file is flushed every `<number>` seconds. If `flush=False` (default),
    the file is flushed only when the file or app is closed.

    This function runs forever, so run it in a separate thread.
    '''

    def __init__(self, **kwargs):
        self.params = kwargs
        self.url = 'https://stream.twitter.com/1.1/statuses/filter.json'
        self.valid_params = {
            'follow',
            'track',
            'locations',
            'delimited',
            'stall_warnings',
            'filter_level',
            'language',
        }
        self.enabled = True
        self.delay = 0

        # Set up writers
        if 'path' in kwargs:
            self.stream = StreamWriter(kwargs['path'], flush=kwargs.get('flush', False))
            self.process_bytes = self.stream.write
        elif 'function' in kwargs:
            self.process_json = build_transform(
                kwargs, vars={'message': {}}, filename='TwitterStream:function'
            )
        elif kwargs.get('driver') == 'sqlalchemy':
            engine = gramex.data.create_engine(kwargs['url'], **kwargs.get('parameters', {}))
            table = gramex.data.get_table(kwargs['table'])
            fields = kwargs['fields']
            for field in list(fields.keys()):
                if field not in table.columns:
                    app_log.error(f'TwitterStream field {field} not in table')
                    fields.pop(field)
            flatten = flattener(fields=fields)
            self.process_json = lambda tweet: engine.execute(table.insert(flatten(tweet)))

        self.buf = bytearray()
        self.client = tornado.httpclient.HTTPClient()
        while True:
            # Set .enabled to False to temporarily disable streamer
            if self.enabled:
                params = {
                    key: val.encode('utf-8')
                    for key, val in self.params.items()
                    if key in self.valid_params
                }
                if 'follow' not in params and 'track' not in params and 'locations' not in params:
                    self.enabled = False
                    self.delay = 5
                    app_log.error('TwitterStream needs follow, track or locations. Disabling')
                else:
                    self.fetch_tweets(params)
            # Restart after a delay determined by
            time.sleep(self.delay)

    def fetch_tweets(self, tweet_params):
        oauth = oauth1.Client(
            client_key=self.params['key'],
            client_secret=self.params['secret'],
            resource_owner_key=self.params['access_key'],
            resource_owner_secret=self.params['access_secret'],
        )
        headers = {
            'Content-Type': 'application/x-www-form-urlencoded',
            'User-Agent': 'Gramex',
        }
        url, headers, data = oauth.sign(
            self.url, 'POST', body=urlencode(tweet_params), headers=headers
        )
        self.req = tornado.httpclient.HTTPRequest(
            method='POST',
            url=url,
            body=data,
            headers=headers,
            request_timeout=10 * 24 * 60 * 60,  # Keep request alive for 10 days
            streaming_callback=self._stream,
            header_callback=self.header_callback,
        )

        try:
            self.headers = None
            self.client.fetch(self.req)
            self.delay = 0
        except tornado.httpclient.HTTPError as e:
            # HTTPError is raised for non-200 HTTP status codes.
            # For rate limiting, start with 1 minute and double each attempt
            if e.code in {RATE_LIMITED, TOO_MANY_REQUESTS}:
                self.delay = self.delay * 2 if self.delay else 60
                app_log.error(
                    f'TwitterStream HTTP {e.code} (rate limited): {e.response}. '
                    f'Retry: {self.delay}s'
                )
            # For Tornado timeout errors, reconnect immediately
            elif e.code == CLIENT_TIMEOUT:
                self.delay = 0
                app_log.error(
                    f'TwitterStream HTTP {e.code} (timeout): {e.response}. '
                    f'Retry: {self.delay}s'
                )
            # For server errors, start with 5 seconds and double until 320 seconds
            elif INTERNAL_SERVER_ERROR <= e.code <= GATEWAY_TIMEOUT:
                self.delay = min(320, self.delay * 2 if self.delay else 1)
                app_log.error(f'TwitterStream HTTP {e.code}: {e.response}. Retry: {self.delay}s')
            # For client errors (e.g. wrong params), disable connection
            else:
                self.delay, self.enabled = 5, False
                app_log.error(f'TwitterStream HTTP {e.code}: {e.response}. Disabling')
        except Exception as e:
            # Other errors are possible, such as IOError.
            # Increase the delay in reconnects by 250ms each attempt, up to 16 seconds.
            self.delay = min(16, self.delay + 0.25)
            app_log.error(f'TwitterStream exception {e}. Retry: {self.delay}s')

    def header_callback(self, line):
        try:
            if self.headers is None:
                start_line = parse_response_start_line(line)
                self.http_version, self.status_code, self.http_reason = start_line
                self.headers = HTTPHeaders()
            else:
                self.headers.parse_line(line)
        except Exception:
            app_log.exception(f'Cannot parse header {line}')

    def _stream(self, data):
        buf = self.buf
        buf.extend(data)
        while len(buf):
            index = buf.find(b'\r\n')
            if index < 0:
                break
            data = bytes(buf[:index])
            del buf[: index + 2]
            # Ignore stall warnings
            if len(data) == 0:
                continue
            try:
                self.process_bytes(data)
            except Exception:
                app_log.exception(f'TwitterStream could not process: {data}')

    def process_bytes(self, data):
        try:
            text = str(data, encoding='utf-8')
            message = json.loads(text)
        except UnicodeError:
            app_log.error(f'TwitterStream unicode error: {data}')
            return
        except ValueError:
            # When rate limited, text="Exceeded connection limit for user"
            app_log.error(f'TwitterStream non-JSON data: {text}')
            return
        # Process the message (which is usually, but not always, a tweet)
        try:
            self.process_json(message)
        except Exception:
            app_log.exception(f'TwitterStream could not process message: {text}')

    def process_json(self, message):
        '''Subclass this to process tweets differently'''
        app_log.info(repr(message))


class StreamWriter:
    def __init__(self, path, flush=False):
        self.path = path
        self.stream = self.stream_path = self.flush_on_write = None
        if isinstance(flush, bool):
            self.flush_on_write = flush
        elif isinstance(flush, (int, float)):
            self.flush_loop = tornado.ioloop.PeriodicCallback(self.flush, flush * 1000)
            self.flush_loop.start()
        else:
            raise ValueError(f'flush={flush!r} is not int/bool')
        self.rotate()

    def flush(self):
        if self.stream is not None:
            self.stream.flush()

    def rotate(self):
        '''
        Create and rotate file streams.

        The `path` format string determines the filename. For example,
        `tweets.{:%Y-%m-%d}.jsonl` creates a filename based on the current
        date, e.g. `tweets.2016-12-31.jsonl`. When rotating, if the new
        filename is the same as the old, the file continues. If it's a different
        file, the old file is closed and the new file is created.

        The rotation frequency is based on the crontab entries in the config,
        i.e. based on `hours`, `days`, `weeks`, etc. It defaults to every
        minute.
        '''
        # First, flush the stream to ensure that data is not lost.
        # Then set up new stream (if required, based on the filename)
        self.flush()
        path = self.path.format(datetime.datetime.utcnow())
        if path != self.stream_path:
            if self.stream is not None:
                self.stream.close()
            self.stream_path = path
            folder = os.path.dirname(os.path.abspath(path))
            if not os.path.exists(folder):
                os.makedirs(folder)
            self.stream = open(path, 'ab')  # noqa SIM115
            app_log.debug(f'StreamWriter writing to {path}')

        # Schedule the next call after a minute
        ioloop = gramex.service.info.main_ioloop or tornado.ioloop.IOLoop.current()
        ioloop.call_later(60, self.rotate)

    def write(self, data):
        self.stream.write(data)
        self.stream.write('\n')
        if self.flush_on_write:
            self.stream.flush()



import json
from .basehandler import BaseWebSocketHandler
from gramex.transforms import build_transform
from tornado.gen import coroutine
from tornado.httpclient import AsyncHTTPClient
from typing import Callable, Dict, Union


class ChatGPTHandler(BaseWebSocketHandler):
    '''A WebSocketHandler that proxies requests to OpenAI chat completions.

    Configuration:

    ```yaml
        handler: ChatGPTHandler
        kwargs:
          key: ...
          max_history: ...
          model: gpt-3.5-turbo
          temperature:
              function: handler.args.get('temperature', 0.9)
          ...
    ```
    '''

    _transform_vars = {
        'open': {'params': None, 'handler': None},
        'prepare': {'message': None, 'handler': None},
        'modify': {'data': None, 'handler': None},
    }
    _api_conf_defaults = {
        'url': 'https://api.openai.com/v1/chat/completions',
        'key': None,
        'max_history': None,
    }
    _params_defaults = {
        'model': 'gpt-3.5-turbo',
        'temperature': None,
        'top_p': None,
        'n': None,
        'stream': {'function': '"stream" in handler.args'},
        'stop': None,
        'max_tokens': None,
        'frequency_penalty': None,
        'presence_penalty': None,
        'logit_bias': None,
        'user': None,
    }

    @classmethod
    def setup(cls, **kwargs):
        super(BaseWebSocketHandler, cls).setup(**kwargs)

        cls._api_conf, cls._params, cls._transform = {}, {}, {}
        # url, key, max_history can be set as key: xxx or key: {function: handler.args.get('key')}
        for k, v in cls._api_conf_defaults.items():
            cls._set_default(k, kwargs.pop(k, v), lambda x: cls._api_conf.update({k: x}))
        # All API params can be set as as key: xxx or key: {function: handler.args.get('key')}
        for k, v in cls._params_defaults.items():
            cls._set_default(k, kwargs.pop(k, v), lambda x: cls._params.update({k: x}))
        # open, prepare and modify are specified directly as the functions
        for k, vars in cls._transform_vars.items():
            v = {'function': kwargs[k]} if k in kwargs else ''
            cls._set_default(k, v, lambda x: cls._transform.update({k: x}), vars=vars)

    @classmethod
    def _set_default(
        cls,
        k: str,
        v: Union[int, str, Callable, Dict[str, str]],
        setter: Callable,
        vars: Dict = {'handler': None},
    ):
        if isinstance(v, dict) and 'function' in v:
            v = build_transform(v, filename=f'url:{cls.name}.{k}', vars=vars, iter=False)
        if v is not None:
            setter(v)

    def open(self):
        self.http = AsyncHTTPClient()
        self.api_conf = {k: v(self) if callable(v) else v for k, v in self._api_conf.items()}
        self.params = {k: v(self) if callable(v) else v for k, v in self._params.items()}
        self.params.setdefault('messages', [])
        self.headers = {'Content-Type': 'application/json'}
        if 'key' in self.api_conf:
            self.headers['Authorization'] = f'Bearer {self.api_conf.get("key", "")}'
        if callable(self._transform['open']):
            self._transform['open'](self.params, self)

    @coroutine
    def on_message(self, message: str):
        if 'max_history' in self.api_conf:
            max_history = self.api_conf['max_history']
            self.params['messages'] = self.params['messages'][-max_history * 2 :]

        if callable(self._transform['prepare']):
            result = self._transform['prepare'](message=message, handler=self)
            if result is not None:
                message = result
        self.params['messages'].append({'role': 'user', 'content': message})

        kwargs = {}
        if self.params['stream']:
            kwargs['streaming_callback'] = self.on_chunk
            self.chunks, self.tokens = [], []
        try:
            r = yield self.http.fetch(
                self.api_conf['url'],
                method='POST',
                request_timeout=0,  # no timeout
                body=json.dumps(self.params),
                headers=self.headers,
                raise_error=False,
                **kwargs,
            )
        except Exception as e:  # noqa - catch all exceptions
            # If there's a non-HTTP error (e.g. timeout), return it
            self.write_message(f'[ERROR] {e}')
            self.write_message('')
            return
        # If there's a HTTP error, return it
        if r.code != 200:
            error = json.loads(r.body or b''.join(self.chunks))['error']
            self.write_message('[ERROR]\n' + '\n'.join(f'{k}: {v}' for k, v in error.items()))
            self.write_message('')
        # If streaming is enabled, ignore the empty r.body. self.on_chunk will handle it
        # If streaming is disabled, return the entire response
        elif r.body:
            data = json.loads(r.body)
            # TODO: Handle multiple responses in data['choices'], i.e. n > 1, properly
            content = data['choices'][0]['message']['content']
            # Call modify(data). Use the return value (if any) as the content
            if callable(self._transform['modify']):
                result = self._transform['modify'](data=data, handler=self)
                if result is not None:
                    content = result
            self.params['messages'].append({'role': 'assistant', 'content': content})
            self.write_message(content)
            self.write_message('')

    def on_chunk(self, chunk: bytes):
        self.chunks.append(chunk.decode('utf-8'))
        [*lines, remaining] = ''.join(self.chunks).split('\n\n')
        self.chunks = [remaining]
        # Each chunk has a data: ... but the last chunk has an additional line with
        # data: [DONE]. We ignore that. Just take the first line with data:
        for line in lines:
            # Ignore empty lines
            if not line.startswith('data: '):
                continue
            # When the last chunk is received, send the entire message
            if line.startswith('data: [DONE]'):
                content = ''.join(self.tokens)
                self.params['messages'].append({'role': 'assistant', 'content': content})
                self.write_message('')
                self.tokens = []
                continue
            data = json.loads(line[6:])
            # TODO: Handle multiple responses in data['choices'], i.e. n > 1, properly
            delta = data['choices'][0]['delta']
            # Not all deltas have content. Starting and ending deltas don't. Skip them
            if 'content' in delta:
                token = delta['content']
                self.tokens.append(token)
                self.write_message(token)



import json
import tornado.web
import tornado.gen
from types import GeneratorType
from gramex.transforms import build_transform
from gramex.config import app_log, CustomJSONEncoder
from .basehandler import BaseHandler


class FunctionHandler(BaseHandler):
    '''Renders the output of a function when the URL is called via GET or POST.

    - `function`: A Python expression that can use `handler` as a variable.
    - `headers`: HTTP headers to set on the response.
    - `redirect`: URL to redirect to when done, e.g. for calculations without output.

    The function result is converted to a string and rendered.
    You can also yield one or more results. These are written immediately, in order.
    '''

    @classmethod
    def setup(cls, headers={}, **kwargs):
        super(FunctionHandler, cls).setup(**kwargs)
        # Don't use cls.info.function = build_transform(...) -- Python treats it as a method
        cls.info = {}
        cls.info['function'] = build_transform(
            kwargs, vars={'handler': None}, filename=f'url:{cls.name}'
        )
        cls.headers = headers
        cls.post = cls.put = cls.delete = cls.patch = cls.get
        if not kwargs.get('cors'):
            cls.options = cls.get

    @tornado.gen.coroutine
    def get(self, *path_args):
        if self.redirects:
            self.save_redirect_page()

        if 'function' not in self.info:
            raise ValueError(f'Invalid function definition in url:{self.name}')
        result = self.info['function'](handler=self)
        for header_name, header_value in self.headers.items():
            self.set_header(header_name, header_value)

        # Use multipart to check if the respose has multiple parts. Don't
        # flush unless it's multipart. Flushing disables Etag
        multipart = isinstance(result, GeneratorType) or len(result) > 1

        # build_transform results are iterable. Loop through each item
        for item in result:
            # Resolve futures and write the result immediately
            if tornado.concurrent.is_future(item):
                item = yield item
            # To check if item is a numpy object, avoid isinstance(numpy.int8), etc.
            # Importing numpy is slow. Instead, check the class name.
            # Strip trailing numbers (e.g. int8, int16, int32)
            # Strip trailing underscore (e.g. str_, bytes_)
            # Strip leading 'u' (e.g. uint, ulong)
            cls = type(item).__name__.rstrip('0123456789_').lstrip('u')
            if isinstance(item, (bytes, str)):
                self.write(item)
                if multipart:
                    self.flush()
            # Ignore None as a return type
            elif item is None:
                pass
            # Allow ANY type that can be converted by CustomJSONEncoder.
            # This includes JSON types, detected by isinstance(item, ...))
            # and numpy types, detected by cls in (...)
            # and anything with a to_dict, e.g. DataFrames
            elif (
                isinstance(item, (int, float, bool, list, tuple, dict))
                or cls in ('datetime', 'int', 'intc', 'float', 'bool', 'ndarray', 'bytes', 'str')
                or hasattr(item, 'to_dict')
            ):
                self.write(
                    json.dumps(
                        item, separators=(',', ':'), ensure_ascii=True, cls=CustomJSONEncoder
                    )
                )
                if multipart:
                    self.flush()
            else:
                app_log.warning(
                    f'url:{self.name}: FunctionHandler can write scalars/list/dict, '
                    f'not {type(item)}: {item!r}'
                )

        if self.redirects:
            self.redirect_next()
